from typing import List

from ..sink_config import SinkConfigBase
from ..sink_base_params import SinkBaseParams


class OpsGenieSinkParams(SinkBaseParams):
    api_key: str
    teams: List[str] = []
    tags: List[str] = []


class OpsGenieSinkConfigWrapper(SinkConfigBase):
    opsgenie_sink: OpsGenieSinkParams

    def get_params(self) -> SinkBaseParams:
        return self.opsgenie_sink




import traceback
import typer
from slack_sdk import WebClient
from slack_sdk.errors import SlackApiError

SLACK_WELCOME_MESSAGE_TITLE = ":large_green_circle: INFO - Welcome to Robusta"
SLACK_WELCOME_MESSAGE_HEADER_WITH_CLUSTER_NAME = (
    "You have signed up for Slack Monitoring on cluster: '{}'."
)
SLACK_WELCOME_THANK_YOU_MESSAGE = "Thank you for using Robusta.dev"
SLACK_WELCOME_SUPPORT_MESSAGE = (
    "If you have any questions or feedback feel free to write us at "
    "<mailto:support@robusta.dev|support@robusta.dev> "
)


def verify_slack_channel(
    slack_api_key: str,
    cluster_name: str,
    channel_name: str,
    workspace: str,
    debug: bool,
) -> bool:
    try:
        output_welcome_message_blocks = __gen_robusta_test_welcome_message(cluster_name)
        slack_client = WebClient(token=slack_api_key)
        slack_client.chat_postMessage(
            channel=channel_name,
            text="Welcome to Robusta",
            blocks=output_welcome_message_blocks,
            display_as_bot=True,
            unfurl_links=True,
            unfurl_media=True,
        )
        return True
    except SlackApiError as e:
        if e.response.data["error"] == "channel_not_found":
            channel_name_styled = typer.style(channel_name, fg=typer.colors.RED)
            typer.secho(
                f"The channel {channel_name_styled} was not found on Slack workspace {workspace}.\n"
                f"Please verify that the channel exists.\n"
                f"If this is a private channel, verify the Robusta app was added to the channel."
                f"(See https://docs.robusta.dev/master/catalog/sinks/slack.html#sending-robusta-notifications-to-a-private-channel)"
            )
        return False
    except Exception as e:
        if debug:
            typer.secho(traceback.format_exc())
    typer.secho(
        f"There was an unknown exception setting up Slack, please contact Robusta support.",
        fg=typer.colors.RED,
    )
    return False


def __gen_robusta_test_welcome_message(cluster_name: str):
    return [
        {
            "type": "header",
            "text": {
                "type": "plain_text",
                "text": SLACK_WELCOME_MESSAGE_TITLE,
                "emoji": True,
            },
        },
        {
            "type": "header",
            "text": {
                "type": "plain_text",
                "text": SLACK_WELCOME_MESSAGE_HEADER_WITH_CLUSTER_NAME.format(
                    cluster_name
                ),
                "emoji": True,
            },
        },
        {
            "type": "section",
            "text": {"type": "mrkdwn", "text": SLACK_WELCOME_THANK_YOU_MESSAGE},
        },
        {
            "type": "section",
            "text": {"type": "mrkdwn", "text": SLACK_WELCOME_SUPPORT_MESSAGE},
        },
    ]



from typing import List

from .msteams_base import MsTeamsBase


class MsTeamsAction(MsTeamsBase):
    def __init__(self, title: str, visible_keys: List[str], invisible_keys: List[str]):
        super().__init__(self.__to_action(title, visible_keys, invisible_keys))

    @classmethod
    def __to_action(cls, title: str, visible_keys: List[str], invisible_keys: List[str]):
        return {
            "selectAction": {
                "type": "Action.ToggleVisibility",
                "title": title,
                "targetElements": cls.__action_toggle_target_elements(visible_keys, invisible_keys),
            }
        }

    @classmethod
    def __action_toggle_target_elements(cls, visible_keys: List[str], invisible_keys: List[str]) -> List[dict]:
        actions = cls.__set_toggle_action(visible_keys, True)
        actions.extend(cls.__set_toggle_action(invisible_keys, False))
        return actions

    @classmethod
    def __set_toggle_action(cls, keys: List[str], visible: bool) -> List[dict]:
        return [{"elementId": key, "isVisible": visible} for key in keys]



from typing import List, Callable, Optional

from ...core.reporting.base import Finding
from ...core.model.base_params import ProcessParams
from ...integrations.kubernetes.custom_models import Process, RobustaPod
from enum import Enum
from ...core.reporting.blocks import (
    BaseBlock,
    CallbackBlock,
    CallbackChoice,
    MarkdownBlock,
    TableBlock,
)


class ProcessType(Enum):
    PYTHON = "python"
    JAVA = "java"

class ProcessFinder:
    """
    Find the processes in a Kubernetes pod which match certain filters.
    """

    def __init__(
        self, pod: RobustaPod, filters: ProcessParams, process_type: ProcessType
    ):
        if process_type not in {ProcessType.PYTHON, ProcessType.JAVA}:
            raise Exception(f"Unsupported process type: {process_type}")
        self.pod = pod
        self.filters = filters
        self.process_type = process_type
        self.all_processes = pod.get_processes()
        self.matching_processes = self.__get_matches(
            self.all_processes, filters, process_type
        )

    def get_match_or_report_error(
        self, finding: Finding, retrigger_text: str, retrigger_action: Callable, debug_action: Callable
    ) -> Optional[Process]:
        """
        Returns the single-matching process. If more than one process matches, blocks will be added to the Finding
        to report the error and optionally allow re-triggering the action with a chosen process.
        """
        if self.has_exactly_one_match():
            return self.get_exact_match()
        elif len(self.matching_processes) == 0:
            finding.add_enrichment(
                [MarkdownBlock(f"No matching processes. The processes in the pod are:")]
                + self.__get_error_blocks(
                    self.all_processes, retrigger_text, retrigger_action, debug_action
                )
            )
            return None
        elif len(self.matching_processes) > 1:
            finding.add_enrichment(
                [
                    MarkdownBlock(
                        f"More than one matching process. The matching processes are:"
                    )
                ]
                + self.__get_error_blocks(
                    self.matching_processes, retrigger_text, retrigger_action, debug_action
                )
            )
            return None

    def has_exactly_one_match(self) -> bool:
        """
        Returns true when exactly one process matches
        """
        return len(self.matching_processes) == 1

    def get_pids(self) -> List[int]:
        """
         Returns all relevant pids
        """
        return [p.pid for p in self.matching_processes]

    def get_lowest_relevant_pid(self) -> int:
        """
         Returns the lowest pid which is most likely the parent process
        """
        return min([p.pid for p in self.matching_processes])

    def get_exact_match(self) -> Process:
        """
        Returns a process matching this class and throws when there is more than one match
        """
        if len(self.matching_processes) != 1:
            raise Exception("Only one match is expected")
        return self.matching_processes[0]

    @staticmethod
    def __get_matches(
        processes: List[Process], filters: ProcessParams, process_type: ProcessType
    ) -> List[Process]:
        """
        Returns the processes that match a ProcessParams class
        """
        pid_to_process = {p.pid: p for p in processes}

        if filters.pid is None:
            return [
                p
                for p in pid_to_process.values()
                if process_type.value in p.exe
                and filters.process_substring in " ".join(p.cmdline)
            ]

        if filters.pid not in pid_to_process:
            return []

        return [pid_to_process[filters.pid]]

    def __get_error_blocks(
        self, processes: List[Process], text: str, action: Callable, debug_action: Callable
    ) -> List[BaseBlock]:
        blocks = [
            TableBlock(
                [[p.pid, p.exe, " ".join(p.cmdline)] for p in processes],
                ["pid", "exe", "cmdline"],
            ),
        ]
        if self.filters.interactive:
            choices = {}
            for proc in processes:
                updated_params = self.filters.copy()
                updated_params.process_substring = ""
                updated_params.pid = proc.pid
                choices[f"{text} {proc.pid}"] = CallbackChoice(
                    action=action,
                    action_params=updated_params,
                    kubernetes_object=self.pod,
                )
            choices[f"Still can't choose?"] = CallbackChoice(
                action=debug_action,
                action_params=self.filters,
                kubernetes_object=self.pod,
            )
            blocks.append(CallbackBlock(choices))
            blocks.append(
                MarkdownBlock(
                    "*After clicking a button please wait up to 120 seconds for a response*"
                )
            )
        return blocks



import logging
import uuid
import re
from enum import Enum
from pydantic.main import BaseModel
from typing import List, Dict
from ..model.env_vars import ROBUSTA_UI_DOMAIN
from ..reporting.consts import FindingSubjectType, FindingSource, FindingType
from ...core.discovery.top_service_resolver import TopServiceResolver


class BaseBlock(BaseModel):
    hidden: bool = False


class FindingSeverity(Enum):
    DEBUG = 0
    INFO = 1
    LOW = 2
    MEDIUM = 3
    HIGH = 4


class Enrichment:
    # These is the actual enrichment data
    blocks: List[BaseBlock] = []
    # General purpose rendering flags, that can be used by specific sinks
    annotations: Dict[str, str] = {}

    def __init__(self, blocks: List[BaseBlock], annotations=None):
        if annotations is None:
            annotations = {}
        self.blocks = blocks
        self.annotations = annotations

    def __str__(self):
        return f"annotations: {self.annotations} Enrichment: {self.blocks} "


class Filterable:
    @property
    def attribute_map(self) -> Dict[str, str]:
        raise NotImplementedError

    def get_invalid_attributes(self, attributes: List[str]) -> List:
        return list(set(attributes) - set(self.attribute_map))

    def attribute_matches(self, attribute: str, expression: str) -> bool:
        value = self.attribute_map[attribute]
        return bool(re.match(expression, value))

    def matches(self, requirements: Dict[str, str]) -> bool:
        invalid_attributes = self.get_invalid_attributes(requirements.keys())
        if len(invalid_attributes) > 0:
            logging.warning(f"Invalid match attributes: {invalid_attributes}")
            return False

        for attribute, expression in requirements.items():
            if not self.attribute_matches(attribute, expression):
                return False
        return True


class FindingSubject:
    def __init__(
        self,
        name: str = None,
        subject_type: FindingSubjectType = FindingSubjectType.TYPE_NONE,
        namespace: str = None,
        node: str = None,
    ):
        self.name = name
        self.subject_type = subject_type
        self.namespace = namespace
        self.node = node


class Finding(Filterable):
    """
    A Finding represents an event that should be sent to sinks.
    """

    def __init__(
        self,
        title: str,
        aggregation_key: str,
        severity: FindingSeverity = FindingSeverity.INFO,
        source: FindingSource = FindingSource.NONE,
        description: str = None,
        subject: FindingSubject = FindingSubject(),
        finding_type: FindingType = FindingType.ISSUE,
        failure: bool = True,
        creation_date: str = None,
    ) -> None:
        self.id: uuid = uuid.uuid4()
        self.title = title
        self.finding_type = finding_type
        self.failure = failure
        self.description = description
        self.source = source
        self.aggregation_key = aggregation_key
        self.severity = severity
        self.category = None  # TODO fill real category
        self.subject = subject
        self.enrichments: List[Enrichment] = []
        self.service_key = TopServiceResolver.guess_service_key(
            name=subject.name, namespace=subject.namespace
        )
        uri_path = (
            f"services/{self.service_key}?tab=grouped" if self.service_key else "graphs"
        )
        self.investigate_uri = f"{ROBUSTA_UI_DOMAIN}/{uri_path}"
        self.creation_date = creation_date

    @property
    def attribute_map(self) -> Dict[str, str]:
        return {
            "title": str(self.title),
            "identifier": str(self.aggregation_key),
            "severity": str(self.severity.name),
            "source": str(self.source.name),
            "type": str(self.finding_type.name),
            "kind": str(self.subject.subject_type.value),
            "namespace": str(self.subject.namespace),
            "node": str(self.subject.node),
            "name": str(self.subject.name),
        }

    def add_enrichment(self, enrichment_blocks: List[BaseBlock], annotations=None):
        if not enrichment_blocks:
            return
        if annotations is None:
            annotations = {}
        self.enrichments.append(Enrichment(enrichment_blocks, annotations))

    def __str__(self):
        return f"title: {self.title} desc: {self.description} severity: {self.severity} sub-name: {self.subject.name} sub-type:{self.subject.subject_type.value} enrich: {self.enrichments}"



from .sink_base_params import SinkBaseParams
from ...core.reporting.base import Finding


class SinkBase:
    def __init__(self, sink_params: SinkBaseParams, registry):
        self.sink_name = sink_params.name
        self.params = sink_params
        self.default = sink_params.default
        self.registry = registry
        global_config = self.registry.get_global_config()

        self.account_id = global_config.get("account_id", "")
        self.cluster_name = global_config.get("cluster_name", "")
        self.signing_key = global_config.get("signing_key", "")

    def stop(self):
        pass

    def accepts(self, finding: Finding) -> bool:
        return finding.matches(self.params.match)

    def write_finding(self, finding: Finding, platform_enabled: bool):
        raise NotImplementedError(
            f"write_finding not implemented for sink {self.sink_name}"
        )



import abc
from pydantic import BaseModel

from .sink_base_params import SinkBaseParams


class SinkConfigBase(BaseModel):

    def get_name(self) -> str:
        """get sink name"""
        return self.get_params().name

    @abc.abstractmethod
    def get_params(self) -> SinkBaseParams:
        """get sink params"""



from typing import Optional

from pydantic import BaseModel

from robusta.core.triggers.error_event_trigger import (
    WarningEventTrigger,
    WarningEventCreateTrigger,
    WarningEventUpdateTrigger,
    WarningEventDeleteTrigger,
)


class CustomTriggers(BaseModel):

    on_kubernetes_warning_event: Optional[WarningEventTrigger]
    on_kubernetes_warning_event_create: Optional[WarningEventCreateTrigger]
    on_kubernetes_warning_event_update: Optional[WarningEventUpdateTrigger]
    on_kubernetes_warning_event_delete: Optional[WarningEventDeleteTrigger]



import logging
import uuid
from collections import defaultdict
from enum import Enum
from typing import List, Optional, Dict, Any
from dataclasses import dataclass, field

from pydantic import BaseModel

from ...integrations.scheduled.playbook_scheduler import PlaybooksScheduler
from ..reporting.base import Finding, BaseBlock


class EventType(Enum):
    KUBERNETES_TOPOLOGY_CHANGE = 1
    PROMETHEUS = 2
    MANUAL_TRIGGER = 3
    SCHEDULED_TRIGGER = 4


class ExecutionEventBaseParams(BaseModel):
    named_sinks: Optional[List[str]] = None


class ExecutionContext(BaseModel):
    account_id: str
    cluster_name: str


# Right now:
# 1. this is a dataclass but we need to make all fields optional in subclasses because of https://stackoverflow.com/questions/51575931/
# 2. this can't be a pydantic BaseModel because of various pydantic bugs (see https://github.com/samuelcolvin/pydantic/pull/2557)
# once the pydantic PR that addresses those issues is merged, this should be a pydantic class
# (note that we need to integrate with dataclasses because of hikaru)
@dataclass
class ExecutionBaseEvent:
    # Collection of findings that should be sent to each sink.
    # This collection is shared between different playbooks that are triggered by the same event.
    sink_findings: Dict[str, List[Finding]] = field(
        default_factory=lambda: defaultdict(list)
    )
    # Target sinks for this execution event. Each playbook may have a different list of target sinks.
    named_sinks: Optional[List[str]] = None
    response: Dict[
        str, Any
    ] = None  # Response returned to caller. For admission or manual triggers for example
    stop_processing: bool = False
    _scheduler: Optional[PlaybooksScheduler] = None
    _context: Optional[ExecutionContext] = None

    def set_context(self, context: ExecutionContext):
        self._context = context

    def get_context(self) -> ExecutionContext:
        return self._context

    def set_scheduler(self, scheduler: PlaybooksScheduler):
        self._scheduler = scheduler

    def get_scheduler(self) -> PlaybooksScheduler:
        return self._scheduler

    def create_default_finding(self) -> Finding:
        """Create finding default fields according to the event type"""
        return Finding(
            title="Robusta notification", aggregation_key="Generic finding key"
        )

    def add_enrichment(
        self,
        enrichment_blocks: List[BaseBlock],
        annotations=None,
    ):
        finding_id: uuid = uuid.uuid4()
        for sink in self.named_sinks:
            if len(self.sink_findings[sink]) == 0:
                sink_finding = self.create_default_finding()
                sink_finding.id = (
                    finding_id  # share the same finding id between different sinks
                )
                self.sink_findings[sink].append(sink_finding)

            self.sink_findings[sink][0].add_enrichment(enrichment_blocks, annotations)

    def add_finding(self, finding: Finding):
        for sink in self.named_sinks:
            if len(self.sink_findings[sink]) > 0:
                logging.warning(
                    f"Overriding active finding for {sink}. new finding: {finding}"
                )

            self.sink_findings[sink].insert(0, finding)

    @staticmethod
    def from_params(params: ExecutionEventBaseParams) -> Optional["ExecutionBaseEvent"]:
        return ExecutionBaseEvent(named_sinks=params.named_sinks)



import logging
import os
import os.path
import colorlog


def init_logging():
    logging_level = os.environ.get("LOG_LEVEL", "INFO")
    logging_format = "%(log_color)s%(asctime)s.%(msecs)03d %(levelname)-8s %(message)s"
    logging_datefmt = "%Y-%m-%d %H:%M:%S"

    if os.environ.get("ENABLE_COLORED_LOGS", "false").lower() == "true":
        print("setting up colored logging")
        colorlog.basicConfig(
            format=logging_format, level=logging_level, datefmt=logging_datefmt
        )
    else:
        print("setting up regular logging")
        logging.basicConfig(
            format=logging_format, level=logging_level, datefmt=logging_datefmt
        )

    logging.getLogger().setLevel(logging_level)
    for logger_name in ["werkzeug", "telethon"]:
        log = logging.getLogger(logger_name)
        log.setLevel(logging.ERROR)

    logging.info(f"logger initialized using {logging_level} log level")



import logging
from collections import defaultdict
from typing import List, Dict, Optional

from ..core.sinks.robusta.robusta_sink_params import RobustaSinkConfigWrapper, RobustaSinkParams
from ..core.sinks.sink_config import SinkConfigBase
from ..core.sinks.sink_factory import SinkFactory
from ..integrations.scheduled.playbook_scheduler_manager import (
    PlaybooksSchedulerManager,
)
from ..integrations.receiver import ActionRequestReceiver
from .playbook_definition import PlaybookDefinition
from ..utils.function_hashes import get_function_hash
from ..core.playbooks.playbook_utils import merge_global_params
from ..core.sinks.sink_base import SinkBase
from ..core.playbooks.base_trigger import TriggerEvent
from ..core.playbooks.actions_registry import ActionsRegistry
from ..runner.telemetry import Telemetry
from ..core.model.env_vars import RUNNER_VERSION, PROMETHEUS_ENABLED



class SinksRegistry:

    def __init__(self, sinks: Dict[str, SinkBase]):
        self.sinks = sinks
        self.default_sinks = [sink.sink_name for sink in sinks.values() if sink.default]
        if not self.default_sinks:
            logging.warning(
                f"No default sinks defined. By default, actions results are ignored."
            )
        platform_sinks = [sink for sink in sinks.values() if isinstance(sink.params, RobustaSinkParams)]
        self.platform_enabled = len(platform_sinks) > 0

    def get_sink_by_name(self, sink_name: str) -> Optional[SinkBase]:
        return self.sinks.get(sink_name)

    def get_all(self) -> Dict[str, SinkBase]:
        return self.sinks

    @classmethod
    def construct_new_sinks(
        cls,
        new_sinks_config: List[SinkConfigBase],
        existing_sinks: Dict[str, SinkBase],
        registry
    ) -> Dict[str, SinkBase]:
    
        new_sink_names = [sink_config.get_name() for sink_config in new_sinks_config]
        # remove deleted sinks
        deleted_sink_names = [
            sink_name
            for sink_name in existing_sinks.keys()
            if sink_name not in new_sink_names
        ]
        for deleted_sink in deleted_sink_names:
            logging.info(f"Deleting sink {deleted_sink}")
            existing_sinks[deleted_sink].stop()
            del existing_sinks[deleted_sink]

        new_sinks = existing_sinks.copy()
        # create new sinks, or update existing if changed
        for sink_config in new_sinks_config:
            try:
                # temporary workaround to skip the default and unconfigured robusta token
                if (
                    isinstance(sink_config, RobustaSinkConfigWrapper)
                    and sink_config.robusta_sink.token == "<ROBUSTA_ACCOUNT_TOKEN>"
                ):
                    continue
                if sink_config.get_name() not in new_sinks.keys():
                    logging.info(
                        f"Adding {type(sink_config)} sink named {sink_config.get_name()}"
                    )
                    new_sinks[sink_config.get_name()] = SinkFactory.create_sink(
                        sink_config, registry
                    )
                elif (
                    sink_config.get_params() != new_sinks[sink_config.get_name()].params
                ):
                    logging.info(
                        f"Updating {type(sink_config)} sink named {sink_config.get_name()}"
                    )
                    new_sinks[sink_config.get_name()].stop()
                    new_sinks[sink_config.get_name()] = SinkFactory.create_sink(
                        sink_config, registry
                    )
            except Exception as e:
                logging.error(
                    f"Error configuring sink {sink_config.get_name()} of type {type(sink_config)}: {e}"
                )

        return new_sinks


class PlaybooksRegistry:
    def get_playbooks(self, trigger_event: TriggerEvent) -> List[PlaybookDefinition]:
        return []

    def get_default_sinks(self):
        return []

    def get_global_config(self) -> dict:
        return {}


class PlaybooksRegistryImpl(PlaybooksRegistry):
    def __init__(
        self,
        active_playbooks: List[PlaybookDefinition],
        actions_registry: ActionsRegistry,
        global_config: dict,
        default_sinks: List[str],
    ):
        self.default_sinks = default_sinks
        self.triggers_to_playbooks = defaultdict(list)
        self.global_config = global_config
        for playbook_def in active_playbooks:
            # Merge playbooks params with global params and default sinks
            if not playbook_def.sinks:
                playbook_def.sinks = default_sinks.copy()

            for action in playbook_def.get_actions():
                action_def = actions_registry.get_action(action.action_name)
                if not action_def:
                    msg = f"Action {action.action_name} not found. Will not be executed"
                    logging.error(msg)
                    raise Exception(msg)
                action.set_func_hash(get_function_hash(action_def.func))
                if action_def.params_type:  # action has params
                    action.action_params = merge_global_params(
                        global_config, action.action_params
                    )
                    if getattr(action_def.params_type, "pre_deploy_func", None):
                        for trigger in playbook_def.triggers:
                            action_params = action_def.params_type(
                                **action.action_params
                            )
                            action_params.pre_deploy_func(trigger.get())

                # validate that the action can be triggered by all playbooks triggers
                for trigger in playbook_def.triggers:
                    exec_event_type = trigger.get().get_execution_event_type()
                    if not issubclass(exec_event_type, action_def.event_type):
                        msg = f"Action {action_def.action_name} cannot be triggered by {exec_event_type}"
                        logging.error(msg)
                        raise Exception(msg)

            playbook_def.post_init()

            # add the playbook only once for each event.
            playbooks_trigger_events = set(
                [
                    trigger_definition.get().get_trigger_event()
                    for trigger_definition in playbook_def.triggers
                ]
            )
            for event in playbooks_trigger_events:
                self.triggers_to_playbooks[event].append(playbook_def)

    def get_playbooks(self, trigger_event: TriggerEvent) -> List[PlaybookDefinition]:
        return self.triggers_to_playbooks.get(trigger_event.get_event_name(), [])

    def get_default_sinks(self) -> List[str]:
        return self.default_sinks

    def get_global_config(self) -> dict:
        return self.global_config


class Registry:
    _actions: ActionsRegistry = ActionsRegistry()
    _playbooks: PlaybooksRegistry = PlaybooksRegistry()
    _sinks: SinksRegistry = None
    _scheduler = None
    _receiver: ActionRequestReceiver = None
    _global_config = dict()
    _telemetry: Telemetry = Telemetry(
        runner_version=RUNNER_VERSION,
        prometheus_enabled=PROMETHEUS_ENABLED, 
    )

    def set_actions(self, actions: ActionsRegistry):
        self._actions = actions

    def get_actions(self) -> ActionsRegistry:
        return self._actions

    def set_playbooks(self, playbooks: PlaybooksRegistry):
        self._playbooks = playbooks

    def get_playbooks(self) -> PlaybooksRegistry:
        return self._playbooks

    def set_sinks(self, sinks: SinksRegistry):
        self._sinks = sinks

    def get_sinks(self) -> SinksRegistry:
        return self._sinks

    def set_scheduler(self, scheduler: PlaybooksSchedulerManager):
        self._scheduler = scheduler

    def get_scheduler(self) -> PlaybooksSchedulerManager:
        return self._scheduler

    def set_receiver(self, receiver: ActionRequestReceiver):
        self._receiver = receiver

    def get_receiver(self) -> ActionRequestReceiver:
        return self._receiver
    
    def get_telemetry(self) -> Telemetry:
        return self._telemetry

    def set_global_config(self, config : Dict ):
        self.global_config = config
    
    def get_global_config(self) -> Dict:
        return self.global_config



from typing import Optional

from dataclasses import dataclass
from hikaru.meta import HikaruDocumentBase

from ...core.reporting import Finding
from ...core.model.k8s_operation_type import K8sOperationType
from ...core.model.events import ExecutionBaseEvent


@dataclass
class K8sBaseChangeEvent(ExecutionBaseEvent):
    description: str = ""
    operation: Optional[
        K8sOperationType
    ] = None  # because this dataclass needs to have defaults :(
    obj: Optional[
        HikaruDocumentBase
    ] = None  # marked as optional because this dataclass needs to have defaults :(
    old_obj: Optional[HikaruDocumentBase] = None  # same above

    def create_default_finding(self) -> Finding:
        if (
            self.obj
            and hasattr(self.obj, "metadata")
            and hasattr(self.obj.metadata, "name")
        ):
            # hikaru's docs say `kind` always exists on HikaruDocumentBase, but it isn't clear from hikaru's code
            kind = getattr(self.obj, "kind", "").lower()
            title = f"{self.operation.value.capitalize()} to {kind} {self.obj.metadata.name}"
        else:
            title = f"Kubernetes {self.operation.value.lower()}"

        return Finding(
            title=title,
            aggregation_key=title,
        )



from typing import List, Optional, Union, Dict
from pydantic import BaseModel, SecretStr, validator

from ..playbooks.playbook_utils import replace_env_vars_values
from ..sinks.webhook.webhook_sink_params import WebhookSinkConfigWrapper
from ..sinks.telegram.telegram_sink_params import TelegramSinkConfigWrapper
from ...model.playbook_definition import PlaybookDefinition
from ..sinks.datadog.datadog_sink_params import DataDogSinkConfigWrapper
from ..sinks.kafka.kafka_sink_params import KafkaSinkConfigWrapper
from ..sinks.msteams.msteams_sink_params import MsTeamsSinkConfigWrapper
from ..sinks.robusta.robusta_sink_params import RobustaSinkConfigWrapper
from ..sinks.slack.slack_sink_params import SlackSinkConfigWrapper
from ..sinks.opsgenie.opsgenie_sink_params import OpsGenieSinkConfigWrapper


class PlaybookRepo(BaseModel):
    url: str
    key: Optional[SecretStr] = SecretStr("")
    pip_install: bool = (
        True  # Set to False, if the playbooks package is already in site-packages.
    )


class RunnerConfig(BaseModel):
    playbook_repos: Dict[str, PlaybookRepo]
    sinks_config: Optional[
        List[
            Union[
                RobustaSinkConfigWrapper,
                SlackSinkConfigWrapper,
                DataDogSinkConfigWrapper,
                KafkaSinkConfigWrapper,
                MsTeamsSinkConfigWrapper,
                OpsGenieSinkConfigWrapper,
                TelegramSinkConfigWrapper,
                WebhookSinkConfigWrapper,
            ]
        ]
    ]
    global_config: Optional[dict] = {}
    active_playbooks: Optional[List[PlaybookDefinition]] = []

    @validator('global_config')
    def env_var_params(cls, global_config: dict):
        return replace_env_vars_values(global_config)



from pydantic import BaseModel, root_validator
from typing import Dict

from robusta.core.playbooks.playbook_utils import replace_env_vars_values


class SinkBaseParams(BaseModel):
    name: str
    default: bool = True
    match: dict = {}

    @root_validator
    def env_values_validation(cls, values: Dict):
        return replace_env_vars_values(values)



from collections import defaultdict
from pydantic.main import BaseModel, Optional
from typing import Dict

class SinkInfo(BaseModel):
    type: str = 'None'
    findings_count: int = 0

class Telemetry(BaseModel):
    account_id: str = ""
    cluster_id: str = ""
    runner_version : str
    last_alert_at : Optional[str] #ts

    nodes_count: int = 0
    playbooks_count : int = 0 
    prometheus_enabled: bool = False
    sinks_info: Dict[str, SinkInfo] = defaultdict(lambda: SinkInfo()) # refresh daily, sinks id -> sink info




from abc import abstractmethod
from typing import Optional, Dict, Any, List, Protocol

from ...model.playbook_action import PlaybookAction
from ...core.model.events import ExecutionBaseEvent
from ...core.playbooks.base_trigger import TriggerEvent
from ...runner.telemetry import Telemetry


class PlaybooksEventHandler(Protocol):
    """Interface for handling trigger events and running playbook actions"""

    @abstractmethod
    def handle_trigger(self, trigger_event: TriggerEvent) -> Optional[Dict[str, Any]]:
        """Handle trigger event. Find the matching playbooks and run their actions"""
        pass

    @abstractmethod
    def run_actions(
        self,
        execution_event: ExecutionBaseEvent,
        actions: List[PlaybookAction],
        sync_response: bool = False,
        no_sinks: bool = False,
    ) -> Optional[Dict[str, Any]]:
        """Run list of actions using the provided execution event"""
        pass

    @abstractmethod
    def run_external_action(
        self,
        action_name: str,
        action_params: Optional[dict],
        sinks: Optional[List[str]],
        sync_response: bool = False,
        no_sinks: bool = False
    ) -> Optional[Dict[str, Any]]:
        """Execute an external action"""
        pass

    @abstractmethod
    def get_global_config(
        self,
    ) -> dict:
        """Return runner global config"""
        pass

    @abstractmethod
    def get_telemetry(
        self,
    ) -> Telemetry:
        """Return runner telemetry"""
        pass



from typing import Union, List, Callable

from ...core.schedule.model import JobState, FixedDelayRepeat, DynamicDelayRepeat


class PlaybooksScheduler:
    def schedule_action(
        self,
        action_func: Callable,
        task_id: str,
        scheduling_params: Union[FixedDelayRepeat, DynamicDelayRepeat],
        named_sinks: List[str],
        action_params=None,
        job_state: JobState = JobState(),
        replace_existing: bool = False,
        standalone_task: bool = False,
    ):
        """Schedule new action job"""
        pass



import importlib.util
import logging
import os
import pkgutil
import subprocess
import sys
import threading
import yaml
from typing import Optional, List, Dict
from inspect import getmembers

from ..cli.utils import get_package_name
from ..integrations.receiver import ActionRequestReceiver
from ..integrations.scheduled.trigger import ScheduledTriggerEvent
from ..core.playbooks.playbooks_event_handler import PlaybooksEventHandler
from ..core.model.runner_config import RunnerConfig, PlaybookRepo
from ..core.playbooks.actions_registry import ActionsRegistry, Action
from ..core.model.env_vars import (
    INTERNAL_PLAYBOOKS_ROOT,
    PLAYBOOKS_CONFIG_FILE_PATH,
    PLAYBOOKS_ROOT, DEFAULT_PLAYBOOKS_ROOT, CUSTOM_PLAYBOOKS_ROOT,
)
from ..integrations.git.git_repo import (
    GitRepoManager,
    GitRepo,
    GIT_SSH_PREFIX,
    GIT_HTTPS_PREFIX,
    LOCAL_PATH_URL_PREFIX,
)
from ..utils.file_system_watcher import FileSystemWatcher
from ..model.playbook_definition import PlaybookDefinition
from ..model.config import (
    Registry,
    SinksRegistry,
    PlaybooksRegistryImpl,
    PlaybooksRegistry,
)
from ..integrations.scheduled.playbook_scheduler_manager_impl import (
    PlaybooksSchedulerManagerImpl,
)
import hashlib

class ConfigLoader:

    # the structure on disk is:
    # root_playbook_path/
    # |- playbook_dir1
    #    |--- playbook1.py
    #    |--- playbook2.py
    #    |--- requirements.txt
    # |- playbook_dir2
    #    |--- ...
    def __init__(
        self,
        registry: Registry,
        event_handler: PlaybooksEventHandler,
    ):
        self.config_file_path = PLAYBOOKS_CONFIG_FILE_PATH
        self.registry = registry
        self.event_handler = event_handler
        self.root_playbook_path = PLAYBOOKS_ROOT
        self.reload_lock = threading.RLock()
        self.watcher = FileSystemWatcher(
            self.root_playbook_path, self.__reload_playbook_packages
        )
        self.conf_watcher = FileSystemWatcher(
            self.config_file_path, self.__reload_playbook_packages
        )
        self.__reload_playbook_packages("initialization")

    def close(self):
        self.watcher.stop_watcher()
        self.conf_watcher.stop_watcher()

    def reload(self, description: str):
        self.__reload_playbook_packages(description)

    def __reload_scheduler(self, playbooks_registry: PlaybooksRegistry):
        scheduler = self.registry.get_scheduler()
        if not scheduler:  # no scheduler yet, initialization
            scheduler = PlaybooksSchedulerManagerImpl(event_handler=self.event_handler)
            self.registry.set_scheduler(scheduler)

        scheduler.update(playbooks_registry.get_playbooks(ScheduledTriggerEvent()))

    def __reload_receiver(self):
        receiver = self.registry.get_receiver()
        if not receiver:  # no existing receiver, just start one
            self.registry.set_receiver(ActionRequestReceiver(self.event_handler))
            return

        current_account_id = self.event_handler.get_global_config().get("account_id")
        current_cluster_name = self.event_handler.get_global_config().get(
            "cluster_name"
        )

        if (
            current_account_id != receiver.account_id
            or current_cluster_name != receiver.cluster_name
        ):
            # need to re-create the receiver
            receiver.stop()
            self.registry.set_receiver(ActionRequestReceiver(self.event_handler))

    @classmethod
    def __get_package_name(cls, local_path) -> str:
        package_name = get_package_name(local_path)
        if not package_name:
            raise Exception(f"Illegal playbooks package {local_path}. Package name not found")
        return package_name

    def __load_playbooks_repos(
        self,
        actions_registry: ActionsRegistry,
        playbooks_repos: Dict[str, PlaybookRepo],
    ):
        playbook_packages = []
        for playbook_package, playbooks_repo in playbooks_repos.items():
            try:
                if (
                    playbooks_repo.pip_install
                ):  # skip playbooks that are already in site-packages
                    if playbooks_repo.url.startswith(GIT_SSH_PREFIX) or playbooks_repo.url.startswith(GIT_HTTPS_PREFIX):
                        repo = GitRepo(
                            playbooks_repo.url,
                            playbooks_repo.key.get_secret_value(),
                        )
                        local_path = repo.repo_local_path
                    elif playbooks_repo.url.startswith(LOCAL_PATH_URL_PREFIX):
                        local_path = playbooks_repo.url.replace(
                            LOCAL_PATH_URL_PREFIX, ""
                        )
                    else:
                        raise Exception(
                            f"Illegal playbook repo url {playbooks_repo.url}. "
                            f"Must start with '{GIT_SSH_PREFIX}', '{GIT_HTTPS_PREFIX}' or '{LOCAL_PATH_URL_PREFIX}'"
                        )

                    if not os.path.exists(
                        local_path
                    ):  # in case the repo url was defined before it was actually loaded
                        logging.error(
                            f"Playbooks local path {local_path} does not exist. Skipping"
                        )
                        continue

                    # Adding to pip the playbooks repo from local_path
                    subprocess.check_call(
                        [sys.executable, "-m", "pip", "install", local_path]
                    )
                    playbook_package = self.__get_package_name(local_path=local_path)

                playbook_packages.append(playbook_package)
            except Exception as e:
                logging.error(f"Failed to add playbooks repo {playbook_package}", exc_info=True)

        for package_name in playbook_packages:
            self.__import_playbooks_package(actions_registry, package_name)

    @classmethod
    def __import_playbooks_package(
        cls, actions_registry: ActionsRegistry, package_name: str
    ):
        logging.info(f"Importing actions package {package_name}")
        pkg = importlib.import_module(package_name)
        playbooks_modules = [
            name for _, name, _ in pkgutil.walk_packages(path=pkg.__path__)
        ]
        for playbooks_module in playbooks_modules:
            try:
                module_name = ".".join([package_name, playbooks_module])
                logging.info(f"importing actions from {module_name}")
                m = importlib.import_module(module_name)
                playbook_actions = getmembers(m, Action.is_action)
                for (action_name, action_func) in playbook_actions:
                    actions_registry.add_action(action_func)
            except Exception as e:
                logging.error(f"failed to module {playbooks_module}", exc_info=True)

    def __reload_playbook_packages(self, change_name):
        logging.info(f"Reloading playbook packages due to change on {change_name}")
        with self.reload_lock:
            try:
                runner_config = self.__load_runner_config(self.config_file_path)
                if runner_config is None:
                    return

                self.registry.set_global_config(runner_config.global_config)
                action_registry = ActionsRegistry()
                # reordering playbooks repos, so that the internal and default playbooks will be loaded first
                # It allows to override these, with playbooks loaded afterwards
                playbook_repos: Dict[str, PlaybookRepo] = {}
                playbook_repos[
                    "robusta.core.playbooks.internal"
                ] = PlaybookRepo(url=INTERNAL_PLAYBOOKS_ROOT, pip_install=False)
                # order matters! Loading the default first, allows overriding it if adding package with the same name
                # since python 3.7, iteration order is identical to insertion order, if dict didn't change
                # default playbooks
                playbook_repos[
                    self.__get_package_name(DEFAULT_PLAYBOOKS_ROOT)
                ] = PlaybookRepo(url=f"file://{DEFAULT_PLAYBOOKS_ROOT}")

                for url, repo in runner_config.playbook_repos.items():
                    playbook_repos[url] = repo

                # saving the ordered playbooks repo into runner config
                runner_config.playbook_repos = playbook_repos
                # custom playbooks
                if os.path.exists(CUSTOM_PLAYBOOKS_ROOT):
                    for custom_playbooks_location in os.listdir(CUSTOM_PLAYBOOKS_ROOT):
                        location = os.path.join(CUSTOM_PLAYBOOKS_ROOT, custom_playbooks_location)
                        runner_config.playbook_repos[
                            self.__get_package_name(location)
                        ] = PlaybookRepo(url=f"file://{location}")
                else:
                    logging.info(f"No custom playbooks defined at {CUSTOM_PLAYBOOKS_ROOT}")

                self.__load_playbooks_repos(
                    action_registry, runner_config.playbook_repos
                )

                (sinks_registry, playbooks_registry) = self.__prepare_runtime_config(
                    runner_config, self.registry.get_sinks(), action_registry, self.registry
                )
                # clear git repos, so it would be re-initialized
                GitRepoManager.clear_git_repos()

                self.__reload_scheduler(playbooks_registry)

                self.registry.set_actions(action_registry)
                self.registry.set_playbooks(playbooks_registry)
                self.registry.set_sinks(sinks_registry)

                telemetry = self.registry.get_telemetry()
                telemetry.playbooks_count = len(runner_config.active_playbooks)
                telemetry.account_id = hashlib.sha256(str(runner_config.global_config.get("account_id", "no_account")).encode("utf-8")).hexdigest()
                telemetry.cluster_id = hashlib.sha256(str(runner_config.global_config.get("cluster_name", "no_cluster")).encode("utf-8")).hexdigest()

                self.__reload_receiver()
            except Exception as e:
                logging.error(f"unknown error reloading playbooks. will try again when they next change", exc_info=True)

    @classmethod
    def __prepare_runtime_config(
        cls,
        runner_config: RunnerConfig,
        sinks_registry: SinksRegistry,
        actions_registry: ActionsRegistry,
        registry : Registry
    ) -> (SinksRegistry, PlaybooksRegistry):
        existing_sinks = sinks_registry.get_all() if sinks_registry else {}
        new_sinks = SinksRegistry.construct_new_sinks(
            runner_config.sinks_config, existing_sinks, registry
        )
        sinks_registry = SinksRegistry(new_sinks)

        # TODO we will replace it with a more generic mechanism, as part of the triggers separation task
        # First, we load the internal playbooks, then add the user activated playbooks
        # Order matters. Internal playbooks, should be added first, and run first
        active_playbooks = [
            PlaybookDefinition(
                triggers=[{"on_kubernetes_any_resource_all_changes": {}}],
                actions=[{"cluster_discovery_updates": {}}],
            )
        ]
        active_playbooks.extend(runner_config.active_playbooks)

        playbooks_registry = PlaybooksRegistryImpl(
            active_playbooks,
            actions_registry,
            runner_config.global_config,
            sinks_registry.default_sinks,
        )

        return sinks_registry, playbooks_registry

    @classmethod
    def __load_runner_config(cls, config_file_path) -> Optional[RunnerConfig]:
        if not os.path.exists(config_file_path):
            logging.warning(
                f"config file not found at {config_file_path} - not configuring any playbooks."
            )
            return None

        logging.info(f"loading config {config_file_path}")
        with open(config_file_path) as file:
            yaml_content = yaml.safe_load(file)
            return RunnerConfig(**yaml_content)



from pydantic.main import BaseModel

from ..sink_config import SinkConfigBase
from ..sink_base_params import SinkBaseParams


class RobustaToken(BaseModel):
    store_url: str
    api_key: str
    account_id: str
    email: str
    password: str


class RobustaSinkParams(SinkBaseParams):
    token: str


class RobustaSinkConfigWrapper(SinkConfigBase):
    robusta_sink: RobustaSinkParams

    def get_params(self) -> SinkBaseParams:
        return self.robusta_sink



# This file was autogenerated. Do not edit.

from hikaru.model.rel_1_16.v1 import *
from ...custom_models import RobustaPod,RobustaDeployment,RobustaJob


KIND_TO_MODEL_CLASS = {
    'Pod': RobustaPod,
    'ReplicaSet': ReplicaSet,
    'DaemonSet': DaemonSet,
    'Deployment': RobustaDeployment,
    'StatefulSet': StatefulSet,
    'Service': Service,
    'Event': Event,
    'HorizontalPodAutoscaler': HorizontalPodAutoscaler,
    'Node': Node,
    'ClusterRole': ClusterRole,
    'ClusterRoleBinding': ClusterRoleBinding,
    'Job': RobustaJob,
    'Namespace': Namespace,
    'ServiceAccount': ServiceAccount,
    'PersistentVolume': PersistentVolume
}



import json
import logging
import requests
from typing import List

from .msteams_elements.msteams_base import MsTeamsBase
from .msteams_adaptive_card_files import MsTeamsAdaptiveCardFiles
from .msteams_elements.msteams_table import MsTeamsTable
from .msteams_elements.msteams_text_block import MsTeamsTextBlock
from .msteams_elements.msteams_column import MsTeamsColumn
from .msteams_elements.msteams_card import MsTeamsCard
from .msteams_elements.msteams_images import MsTeamsImages
from ...core.reporting import FileBlock, TableBlock, ListBlock, MarkdownBlock, KubernetesDiffBlock, HeaderBlock


class MsTeamsMsg:
    # actual size according to the DOC is ~28K.
    # it's hard to determine the real size because for example there can be large images that doesn't count
    # and converting the map to json doesn't give us an exact indication of the size so we need to take 
    # a safe zone of less then 28K
    MAX_SIZE_IN_BYTES = (1024 * 20)

    def __init__(self, webhook_url: str):
        self.entire_msg: List[MsTeamsBase] = []
        self.current_section: List[MsTeamsBase] = []
        self.text_file_containers = []
        self.webhook_url = webhook_url

    def write_title_and_desc(self, title: str, description: str, severity: str,
                             platform_enabled: bool, investigate_uri: str, cluster_name: str):
        block = MsTeamsTextBlock(text=f"{severity} - {title}", font_size='extraLarge')
        self.__write_to_entire_msg([block])
        if platform_enabled:  # add link to the Robusta ui, if it's configured
            self.__write_to_entire_msg([MsTeamsTextBlock(text=f"<{investigate_uri}|Investigate>")])

        self.__write_to_entire_msg([MsTeamsTextBlock(text=f"**Source:** *{cluster_name}*")])

        if description is not None:
            block = MsTeamsTextBlock(text=description)
            self.__write_to_entire_msg([block])

    def write_current_section(self):
        if len(self.current_section) == 0:
            return

        space_block = MsTeamsTextBlock(text=' ', font_size='small')
        separator_block = MsTeamsTextBlock(text=' ', separator=True)

        underline_block = MsTeamsColumn()
        underline_block.add_column(items=[space_block, separator_block], width_stretch=True)

        self.__write_to_entire_msg([underline_block])
        self.__write_to_entire_msg(self.current_section)
        self.current_section = []

    def __write_to_entire_msg(self, blocks: List[MsTeamsBase]):
        self.entire_msg += blocks

    def __write_to_current_section(self, blocks: List[MsTeamsBase]):
        self.current_section += blocks

    def __sub_section_separator(self):
        if len(self.current_section) == 0:
            return
        space_block = MsTeamsTextBlock(text=' ', font_size='small')
        separator_block = MsTeamsTextBlock(text='_' * 30, font_size='small', horizontal_alignment='center')
        self.__write_to_current_section([space_block,separator_block,space_block,space_block])

    def upload_files(self, file_blocks: List[FileBlock]):
        msteams_files = MsTeamsAdaptiveCardFiles()
        block_list: List[MsTeamsBase] = msteams_files.upload_files(file_blocks)
        if len(block_list) > 0:
            self.__sub_section_separator()

        self.text_file_containers += msteams_files.get_text_files_containers_list()
        
        self.__write_to_current_section(block_list)

    def table(self, table_block : TableBlock):
        blocks: List[MsTeamsBase] = []
        if table_block.table_name:
            blocks.append(MsTeamsTextBlock(table_block.table_name))
        blocks.append(MsTeamsTable(list(table_block.headers), table_block.rows))
        self.__write_to_current_section(blocks)
    
    def items_list(self, block: ListBlock):
        self.__sub_section_separator()
        for line in block.items:
            bullet_lines = '\n- ' + line + '\n'
            self.__write_to_current_section([MsTeamsTextBlock(bullet_lines)])

    def diff(self, block: KubernetesDiffBlock):
        rows = [f"*{diff.formatted_path}*: {diff.other_value} -> {diff.value}" for diff in block.diffs]

        list_blocks = ListBlock(rows)
        self.items_list(list_blocks)

    def markdown_block(self, block: MarkdownBlock):
        if not block.text:
            return
        self.__write_to_current_section([MsTeamsTextBlock(block.text)])

    def divider_block(self):
        self.__write_to_current_section([MsTeamsTextBlock('\n\n')])

    def header_block(self, block: HeaderBlock):
        current_header_string = block.text + '\n\n'
        self.__write_to_current_section([MsTeamsTextBlock(current_header_string, font_size='large')])

    # dont include the base 64 images in the total size calculation
    def _put_text_files_data_up_to_max_limit(self, complete_card_map: map):
        curr_images_len = 0
        for element in self.entire_msg:
            if isinstance(element, MsTeamsImages):
                curr_images_len += element.get_images_len_in_bytes()

        max_len_left = self.MAX_SIZE_IN_BYTES - (self.__get_current_card_len(complete_card_map) - curr_images_len)

        curr_line = 0
        while True:
            line_added = False
            curr_line += 1
            for text_element, lines in self.text_file_containers:
                if len(lines) < curr_line:
                    continue

                line = lines[len(lines) - curr_line]
                max_len_left -= len(line)
                if max_len_left < 0:
                    return
                new_text_value = line + text_element.get_text_from_block()
                text_element.set_text_from_block(new_text_value)
                line_added = True

            if not line_added:
                return

    def send(self):
        try:
            complete_card_map: dict = MsTeamsCard(self.entire_msg).get_map_value()
            self._put_text_files_data_up_to_max_limit(complete_card_map)

            response = requests.post(self.webhook_url, json=complete_card_map)
            if response.status_code not in [200, 201]:
                logging.error(f"Error sending to ms teams json: {complete_card_map} error: {response.reason}")

            if response.text and "error" in response.text.lower():  # teams error indication is in the text only :(
                logging.error(f"Failed to send message to teams. error: {response.text} message: {complete_card_map}")

        except Exception as e:
            logging.error(f"error sending message to msteams\ne={e}\n")   

    @classmethod
    def __get_current_card_len(cls, complete_card_map: dict):
        return len(json.dumps(complete_card_map, ensure_ascii=True, indent=2))



from typing import List

from .msteams_base import MsTeamsBase


class MsTeamsCard(MsTeamsBase):
    def __init__(self, elements: List[MsTeamsBase]):
        content = {
            "$schema": "http://adaptivecards.io/schemas/adaptive-card.json",
            "type": "AdaptiveCard",
            "version": "1.2",
            "msTeams": {
                "width": "full"
            },
            "body": [elem.get_map_value() for elem in elements]
        }

        attachment = {
            "contentType": "application/vnd.microsoft.card.adaptive",
            "contentUrl": None,
            "content": content
        }

        block = {
            "type": "message",
            "attachments": [attachment]
        }

        super().__init__(block)



import tempfile
import base64
import os
import uuid
from PIL import Image

from ...core.reporting.blocks import *
from .msteams_elements.msteams_images import MsTeamsImages
from cairosvg import svg2png

from ...core.reporting.utils import is_image, file_suffix_match, JPG_SUFFIX, PNG_SUFFIX


class MsTeamsAdaptiveCardFilesImage:

    @classmethod
    def create_files_for_presentation(cls, file_blocks: List[FileBlock]) -> map:
        encoded_images = []
        image_file_blocks = [file_block for file_block in file_blocks if is_image(file_block.filename)]
        for image_file_block in image_file_blocks:
            encoded_images.append(
                cls.__convert_bytes_to_base_64_url(image_file_block.filename, image_file_block.contents)
            )

        if len(encoded_images) == 0:
            return []
        return MsTeamsImages(encoded_images)

    @classmethod
    def __get_tmp_file_path(cls):
        return os.path.join(tempfile.gettempdir(), str(uuid.uuid1()))

    @classmethod
    def __convert_bytes_to_base_64_url(cls, file_name: str, image_bytes: bytes):
        if file_suffix_match(file_name, JPG_SUFFIX):
            return cls.__jpg_convert_bytes_to_base_64_url(image_bytes)
        if file_suffix_match(file_name, PNG_SUFFIX):
            return cls.__png_convert_bytes_to_base_64_url(image_bytes)
        return cls.__svg_convert_bytes_to_jpg(image_bytes)

    @classmethod
    def __jpg_convert_bytes_to_base_64_url(cls, jpg_bytes: bytes):
        b64_string = base64.b64encode(jpg_bytes).decode("utf-8")
        return 'data:image/jpeg;base64,{0}'.format(b64_string)

    # msteams cant read parsing of url to 'data:image/png;base64,...
    @classmethod
    def __png_convert_bytes_to_base_64_url(cls, png_bytes : bytes):
        png_file_path = cls.__get_tmp_file_path() + PNG_SUFFIX
        jpg_file_path = cls.__get_tmp_file_path() + JPG_SUFFIX
        with open(png_file_path, 'wb') as f:
            f.write(png_bytes)

        im = Image.open(png_file_path)
        rgb_im = im.convert('RGB')
        rgb_im.save(jpg_file_path)
        with open(jpg_file_path, 'rb') as f:
            jpg_bytes = f.read()

        os.remove(png_file_path)
        os.remove(jpg_file_path)

        return cls.__jpg_convert_bytes_to_base_64_url(jpg_bytes)

    # msteams cant read parsing of url to svg image
    @classmethod
    def __svg_convert_bytes_to_jpg(cls, svg_bytes : bytes):
        return cls.__png_convert_bytes_to_base_64_url(svg2png(bytestring=svg_bytes))


import logging

import requests


class ArgoCDClient:
    SYNC_APP_PATTERN = "{0}/api/v1/applications/{1}/sync"

    def __init__(self, url: str, token: str, verify_cert: bool = True):
        self.url = url
        self.token = token
        self.verify_cert = verify_cert
        self.headers = {
            "Authorization": f"Bearer {self.token}",
        }

    def sync_application(self, application_name: str) -> bool:
        sync_url = self.SYNC_APP_PATTERN.format(self.url, application_name)

        argo_response = requests.post(
            sync_url, headers=self.headers, verify=self.verify_cert
        )
        if argo_response.status_code != 200:
            logging.error(
                f"Failed to sync application {application_name} on argo {self.url} {argo_response.reason}"
            )
            return False

        return True



from dataclasses import dataclass

from ...core.reporting import Finding
from ...core.model.events import ExecutionBaseEvent


@dataclass
class ScheduledExecutionEvent(ExecutionBaseEvent):
    recurrence: int = 0

    def create_default_finding(self):
        return Finding(
            title="General scheduled task",
            aggregation_key="General scheduled task",
        )



def exact_match(expected, field_value) -> bool:
    if expected is None:
        return True

    return expected == field_value


def prefix_match(prefix: str, field_value: str) -> bool:
    if not prefix:
        return True

    if field_value is None:
        # we have a prefix requirement, but field doesn't exist. no match
        return False

    return field_value.startswith(prefix)



import hashlib
from typing import List, Any, Dict
from pydantic import PrivateAttr, BaseModel

from ..core.playbooks.trigger import Trigger
from .playbook_action import PlaybookAction


class PlaybookDefinition(BaseModel):
    sinks: List[str] = None
    triggers: List[Trigger]
    actions: List[Dict[str, Any]]
    _actions: List[PlaybookAction] = PrivateAttr()
    stop: bool = False
    _playbook_id: str = PrivateAttr()

    def __init__(self, **data: Any):
        super().__init__(**data)
        self._actions = []
        for action in self.actions:
            if len(action.keys()) != 1:
                raise Exception("Action must have a single name")

            (action_name, action_params) = next(iter(action.items()))
            self._actions.append(
                PlaybookAction(action_name=action_name, action_params=action_params)
            )

    def get_actions(self) -> List[PlaybookAction]:
        return self._actions

    def __playbook_hash(self):
        hash_input = (
            ".".join([trigger.get().json() for trigger in self.triggers])
            + ".".join([action.as_str() for action in self.get_actions()])
            + f"{self.sinks}"
            + str(self.stop)
        )
        return hashlib.md5(hash_input.encode()).hexdigest()

    def post_init(self):
        self._playbook_id = self.__playbook_hash()

    def get_id(self) -> str:
        return self._playbook_id



from ..core.model.base_params import *
from ..core.model.pods import *
from ..core.sinks import *
from ..core.sinks.sink_base import *
from ..core.sinks.sink_config import *
from ..core.sinks.kafka import *
from ..core.reporting.consts import *
from ..core.reporting.callbacks import *
from ..core.reporting.custom_rendering import *
from ..core.schedule.model import *
from ..core.model.env_vars import *
from ..core.model.k8s_operation_type import *

from ..core.reporting import *
from ..core.reporting.action_requests import *

from ..integrations.kubernetes.custom_models import *
from ..integrations.kubernetes.autogenerated.events import *
from ..integrations.kubernetes.autogenerated.models import *
from ..integrations.kubernetes.process_utils import ProcessType, ProcessFinder
from ..integrations.prometheus.models import *
from ..integrations.prometheus.utils import *
from ..integrations.slack.sender import *
from ..integrations.grafana import *
from ..integrations.scheduled.playbook_scheduler_manager_impl import *
from ..integrations.git.git_repo import *
from ..integrations.argocd.argocd_client import *
from ..integrations.resource_analysis.node_cpu_analyzer import NodeCpuAnalyzer
from ..integrations.prometheus.utils import PrometheusDiscovery
from ..core.persistency.in_memory import get_persistent_data
from ..utils.parsing import load_json
from ..utils.rate_limiter import RateLimiter
from ..utils.common import *
from ..utils.function_hashes import action_hash
from ..runner.object_updater import *
from ..core.playbooks.trigger import *
from ..core.playbooks.actions_registry import action
from ..integrations.scheduled.trigger import (
    DynamicDelayRepeatTrigger,
    FixedDelayRepeatTrigger,
)
from ..core.playbooks.common import get_resource_events_table
from ..core.reporting.finding_subjects import PodFindingSubject, KubeObjFindingSubject
from ..core.playbooks.prometheus_enrichment_utils import *



import abc
from typing import Optional, Union
from pydantic import BaseModel

from .event import ScheduledExecutionEvent
from ...core.schedule.model import FixedDelayRepeat, DynamicDelayRepeat
from ...core.playbooks.base_trigger import TriggerEvent, BaseTrigger


class ScheduledTriggerEvent(TriggerEvent):
    def get_event_name(self) -> str:
        return ScheduledTriggerEvent.__name__


class ScheduledTrigger(BaseTrigger):
    def get_trigger_event(self):
        return ScheduledTriggerEvent.__name__

    @staticmethod
    def get_execution_event_type() -> type:
        return ScheduledExecutionEvent

    @abc.abstractmethod
    def get_params(self) -> Union[FixedDelayRepeat, DynamicDelayRepeat]:
        """Get scheduling params"""
        pass


class FixedDelayRepeatTrigger(ScheduledTrigger):
    fixed_delay_repeat: Optional[FixedDelayRepeat]

    def get_params(self) -> Union[FixedDelayRepeat, DynamicDelayRepeat]:
        return self.fixed_delay_repeat


class DynamicDelayRepeatTrigger(ScheduledTrigger):
    dynamic_delay_repeat: Optional[DynamicDelayRepeat]

    def get_params(self) -> Union[FixedDelayRepeat, DynamicDelayRepeat]:
        return self.dynamic_delay_repeat


class ScheduledTriggers(BaseModel):
    on_schedule: Optional[Union[FixedDelayRepeatTrigger, DynamicDelayRepeatTrigger]]



from typing import List

from ....core.model.env_vars import TEAMS_IMAGE_WIDTH
from .msteams_base import MsTeamsBase


class MsTeamsImages(MsTeamsBase):
    def __init__(self, encoded_images: List[str]):
        self.images_len_in_bytes = 0
        image_elements = [self.__to_image(img) for img in encoded_images]
        super().__init__(self.__to_image_set(image_elements))

    def get_images_len_in_bytes(self):
        return self.images_len_in_bytes

    @classmethod
    def __to_image_set(cls, image_elements: List[dict]) -> dict:
        return {
                    "type": "ImageSet",
                    "imageSize": "large",
                    "images": image_elements
                }

    def __to_image(self, encoded_image: str) -> dict:
        self.images_len_in_bytes += len(encoded_image)
        return {
            "type": "Image",
            "url": encoded_image,
            "width": TEAMS_IMAGE_WIDTH
        }



from collections import defaultdict

from typing import List

from ..discovery.top_service_resolver import TopServiceResolver
from ...core.playbooks.base_trigger import TriggerEvent
from ...integrations.kubernetes.autogenerated.triggers import (
    EventAllChangesTrigger,
    EventChangeEvent,
)
from ...integrations.kubernetes.base_triggers import K8sTriggerEvent
from ...utils.rate_limiter import RateLimiter


class WarningEventTrigger(EventAllChangesTrigger):
    rate_limit: int = 3600
    operations: List[str] = None

    def __init__(
        self,
        name_prefix: str = None,
        namespace_prefix: str = None,
        labels_selector: str = None,
        rate_limit: int = 3600,
        operations: List[str] = None,
    ):
        super().__init__(
            name_prefix=name_prefix,
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )
        self.rate_limit = rate_limit
        self.operations = operations

    def should_fire(self, event: TriggerEvent, playbook_id: str):
        should_fire = super().should_fire(event, playbook_id)
        if not should_fire:
            return should_fire

        if not isinstance(event, K8sTriggerEvent):
            return False

        exec_event = self.build_execution_event(event, {})

        if not isinstance(exec_event, EventChangeEvent):
            return False

        if not exec_event.obj or not exec_event.obj.involvedObject:
            return False

        if exec_event.get_event().type != "Warning":
            return False

        if self.operations and exec_event.operation.value not in self.operations:
            return False

        # Perform a rate limit for this service key according to the rate_limit parameter
        name = exec_event.obj.involvedObject.name
        namespace = (
            exec_event.obj.involvedObject.namespace
            if exec_event.obj.involvedObject.namespace
            else ""
        )
        service_key = TopServiceResolver.guess_service_key(
            name=name, namespace=namespace
        )
        return RateLimiter.mark_and_test(
            f"WarningEventTrigger_{playbook_id}_{exec_event.obj.reason}",
            service_key if service_key else namespace + ":" + name,
            self.rate_limit,
        )


class WarningEventCreateTrigger(WarningEventTrigger):
    def __init__(
        self,
        name_prefix: str = None,
        namespace_prefix: str = None,
        labels_selector: str = None,
        rate_limit: int = 3600,
    ):
        super().__init__(
            name_prefix=name_prefix,
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
            rate_limit=rate_limit,
            operations=["create"],
        )


class WarningEventUpdateTrigger(WarningEventTrigger):
    def __init__(
        self,
        name_prefix: str = None,
        namespace_prefix: str = None,
        labels_selector: str = None,
        rate_limit: int = 3600,
    ):
        super().__init__(
            name_prefix=name_prefix,
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
            rate_limit=rate_limit,
            operations=["update"],
        )


class WarningEventDeleteTrigger(WarningEventTrigger):
    def __init__(
        self,
        name_prefix: str = None,
        namespace_prefix: str = None,
        labels_selector: str = None,
        rate_limit: int = 3600,
    ):
        super().__init__(
            name_prefix=name_prefix,
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
            rate_limit=rate_limit,
            operations=["delete"],
        )



# TODO: move to utils
import yaml
import logging
import jsonref
from collections import defaultdict
from typing import Type, List, Callable, Optional, Union, get_origin, get_args
from ..playbooks.base_trigger import BaseTrigger, ExecutionBaseEvent
from ..playbooks.actions_registry import Action
from ..playbooks.trigger import Trigger
from ...integrations.kubernetes.autogenerated.events import (
    KubernetesAnyChangeEvent,
    KubernetesResourceEvent,
)
from ...utils.json_schema import example_from_schema


def get_possible_types(t):
    """
    Given a type or a Union of types, returns a list of the actual types
    """
    if get_origin(t) == Union:
        return get_args(t)
    else:
        return [t]


# see https://stackoverflow.com/questions/13518819/avoid-references-in-pyyaml
class NoAliasDumper(yaml.SafeDumper):
    def ignore_aliases(self, data):
        return True


class ExamplesGenerator:

    SUBTRIGGER_MARKER = "..."
    ANY_TRIGGER_MARKER = "any trigger"

    def __init__(self):
        self.events_to_triggers = defaultdict(set)
        self.triggers_to_yaml = {}

        for field_name, field in Trigger.__fields__.items():
            trigger_classes = [
                t for t in get_possible_types(field.type_) if issubclass(t, BaseTrigger)
            ]
            if len(trigger_classes) == 0:
                continue

            for t in trigger_classes:
                self.triggers_to_yaml[t] = field_name
                execution_event = t.get_execution_event_type()
                possible_events = [execution_event] + list(
                    cls
                    for cls in execution_event.__mro__
                    if issubclass(cls, ExecutionBaseEvent)
                )
                for e in possible_events:
                    self.events_to_triggers[e].add(t)

    def get_possible_triggers(self, event_cls: Type[ExecutionBaseEvent]) -> List[str]:
        name = event_cls.__name__
        if name == "ExecutionBaseEvent":
            return ["on_pod_create"]
        triggers = self.events_to_triggers.get(event_cls)
        if triggers is None:
            raise Exception(f"Don't know how to generate an example trigger for {name}")
        return [self.triggers_to_yaml[t] for t in triggers]

    @classmethod
    def get_manual_trigger_cmd(cls, action: Action) -> str:
        action_params_sample = ""
        if action.params_type:
            required: List[str] = action.params_type.schema().get("required", [])
            for field in required:
                action_params_sample = (
                    action_params_sample + f" {field}={field.upper()}"
                )

        if action.event_type.__name__ == "ExecutionBaseEvent":
            return (
                f"robusta playbooks trigger {action.action_name} {action_params_sample}"
            )

        from_params_parameters_class = getattr(
            action, "from_params_parameter_class", None
        )
        if not from_params_parameters_class:
            return ""

        cmd = f"robusta playbooks trigger {action.action_name}"
        from_params_k8s_type = (
            from_params_parameters_class.schema()["title"]
            .replace("Attributes", "")
            .upper()
        )
        required_fields: List[str] = from_params_parameters_class.schema()["required"]
        for field in required_fields:
            cmd = cmd + f" {field}={from_params_k8s_type}_{field.upper()}"

        return cmd + f" {action_params_sample}"

    # build a triggers list for events that much k8s objects
    @classmethod
    def get_k8s_trigger_list(
        cls, name: str, all_triggers, include_prometheus: bool = False
    ):
        if name not in ["KubernetesAnyChangeEvent", "KubernetesResourceEvent"]:
            return all_triggers

        any_triggers = [
            trigger
            for trigger in all_triggers
            if "on_kubernetes_any_resource" in trigger
        ]

        if include_prometheus:
            any_triggers.extend(
                [trigger for trigger in all_triggers if "prometheus" in trigger]
            )

        any_triggers.append(cls.SUBTRIGGER_MARKER)
        return any_triggers

    def get_supported_triggers(self, action: Action) -> List[str]:
        """
        Get supported triggers for docs.
        Return a list (or sublist) of supported triggers, and the manual trigger command, if supported
        """
        event_cls: Type[ExecutionBaseEvent] = action.event_type
        name = event_cls.__name__
        if name == "ExecutionBaseEvent":
            return [self.ANY_TRIGGER_MARKER]

        # remove duplications and sort
        all_triggers = sorted(list(set(self.get_possible_triggers(event_cls))))

        if len(all_triggers) == 1:  # prometheus alert trigger, scheduled trigger
            return list(all_triggers)

        if issubclass(event_cls, KubernetesAnyChangeEvent):  # change trigger
            return self.get_k8s_trigger_list(name, all_triggers)

        # validating we don't have unexpected trigger
        if not issubclass(event_cls, KubernetesResourceEvent):
            logging.error(f"Found unexpected trigger {event_cls}")
            return ["no trigger"]

        # Resource trigger, PodEvent, DeploymentEvent etc
        return self.get_k8s_trigger_list(name, all_triggers, True)

    def generate_example_config(
        self, action_func: Callable, suggested_trigger: Optional[str]
    ):
        action_metadata = Action(action_func)
        if suggested_trigger is not None:
            trigger = suggested_trigger
        else:
            trigger = self.get_possible_triggers(action_metadata.event_type)[0]
        example = {
            "actions": [{action_metadata.action_name: {}}],
            "triggers": [{trigger: {}}],
        }
        if action_metadata.params_type:
            action_model = action_metadata.params_type
            # instead of loading the schema as python object directly with action_model.schema()
            # we dump to json and then read-back from json to python using jsonref
            # this is necessary to fix parts of the schema that refer to one another
            # without it we need to understand json references and handle them ourselves
            action_schema = jsonref.loads(action_model.schema_json())
            action_example = example_from_schema(action_schema)
            example["actions"][0][action_metadata.action_name] = action_example
        return yaml.dump(example, Dumper=NoAliasDumper)



import time
from typing import Type, TypeVar, List, Dict

import hikaru
import json
import yaml
from hikaru.model import *  # *-import is necessary for hikaru subclasses to work
from pydantic import BaseModel

from ...core.model.env_vars import INSTALLATION_NAMESPACE
from .api_client_utils import *
from .templates import get_deployment_yaml

S = TypeVar("S")
T = TypeVar("T")
PYTHON_DEBUGGER_IMAGE = (
    "us-central1-docker.pkg.dev/genuine-flight-317411/devel/debug-toolkit:v4.3"
)
JAVA_DEBUGGER_IMAGE = (
    "us-central1-docker.pkg.dev/genuine-flight-317411/devel/java-toolkit-11:v1"
)

# TODO: import these from the python-tools project
class Process(BaseModel):
    pid: int
    exe: str
    cmdline: List[str]


class ProcessList(BaseModel):
    processes: List[Process]


def _get_match_expression_filter(expression: LabelSelectorRequirement) -> str:
    if expression.operator.lower() == "exists":
        return expression.key
    elif expression.operator.lower() == "doesnotexist":
        return f"!{expression.key}"

    values = ",".join(expression.values)
    return f"{expression.key} {expression.operator} ({values})"


def build_selector_query(selector: LabelSelector) -> str:
    label_filters = [f"{label[0]}={label[1]}" for label in selector.matchLabels.items()]
    label_filters.extend([
        _get_match_expression_filter(expression) for expression in selector.matchExpressions
    ])
    return ",".join(label_filters)


def get_images(containers: List[Container]) -> Dict[str, str]:
    """
    Takes a list of containers and returns a dict mapping image name to image tag.
    """
    name_to_version = {}
    for container in containers:
        if ":" in container.image:
            image_name, image_tag = container.image.split(":", maxsplit=1)
            name_to_version[image_name] = image_tag
        else:
            name_to_version[container.image] = "<NONE>"
    return name_to_version


def extract_images(k8s_obj: HikaruDocumentBase) -> Optional[Dict[str, str]]:
    containers_paths = [
        [
            "spec",
            "template",
            "spec",
            "containers",
        ],  # deployment, replica set, daemon set, stateful set, job
        ["spec", "containers"],  # pod
    ]

    for path in containers_paths:
        try:
            containers = k8s_obj.object_at_path(path)
            if containers:
                return get_images(containers)
        except Exception:  # Path not found on object, not a real error
            pass

    # no containers found on that k8s obj
    return None


def does_daemonset_have_toleration(ds: DaemonSet, toleration_key: str) -> bool:
    return any(t.key == toleration_key for t in ds.spec.template.spec.tolerations)


def does_node_have_taint(node: Node, taint_key: str) -> bool:
    return any(t.key == taint_key for t in node.spec.taints)


class RobustaEvent:
    @classmethod
    def get_events(cls, kind: str, name: str, namespace: str = None) -> EventList:
        field_selector = f"involvedObject.kind={kind},involvedObject.name={name}"
        if namespace:
            field_selector += f",involvedObject.namespace={namespace}"

        return Event.listEventForAllNamespaces(field_selector=field_selector).obj


class RobustaPod(Pod):
    def exec(self, shell_command: str, container: str = None) -> str:
        """Execute a command inside the pod"""
        if container is None:
            container = self.spec.containers[0].name

        return exec_shell_command(
            self.metadata.name, shell_command, self.metadata.namespace, container
        )

    def get_logs(self, container=None, previous=None, tail_lines=None) -> str:
        """
        Fetch pod logs
        """
        if container is None:
            container = self.spec.containers[0].name
        return get_pod_logs(
            self.metadata.name,
            self.metadata.namespace,
            container,
            previous,
            tail_lines,
        )

    @staticmethod
    def exec_in_java_pod(
        pod_name: str, node_name: str, debug_cmd=None, override_jtk_image:str=JAVA_DEBUGGER_IMAGE
    ) -> str:
        return RobustaPod.exec_in_debugger_pod(
            pod_name,
            node_name,
            debug_cmd, debug_image=override_jtk_image)

    @staticmethod
    def create_debugger_pod(
            pod_name: str, node_name: str, debug_image=PYTHON_DEBUGGER_IMAGE, debug_cmd=None,
            env: Optional[List[EnvVar]] = None, mount_host_root: bool = False) -> "RobustaPod":
        """
        Creates a debugging pod with high privileges
        """

        volume_mounts = None
        volumes = None
        if mount_host_root:
            volume_mounts = [VolumeMount(name="host-root", mountPath="/host")]
            volumes = [Volume(
                name="host-root",
                hostPath=HostPathVolumeSource(
                    path="/",
                    type="Directory"
                )
            )]

        debugger = RobustaPod(
            apiVersion="v1",
            kind="Pod",
            metadata=ObjectMeta(
                name=to_kubernetes_name(pod_name, "debug-"),
                namespace=INSTALLATION_NAMESPACE,
            ),
            spec=PodSpec(
                hostPID=True,
                nodeName=node_name,
                restartPolicy="OnFailure",
                containers=[
                    Container(
                        name="debugger",
                        image=debug_image,
                        imagePullPolicy="Always",
                        command=prepare_pod_command(debug_cmd),
                        securityContext=SecurityContext(
                            capabilities=Capabilities(add=["SYS_PTRACE", "SYS_ADMIN"]),
                            privileged=True
                        ),
                        volumeMounts=volume_mounts,
                        env=env
                    )
                ],
                volumes=volumes
            ),
        )
        # TODO: check the result code
        debugger = debugger.createNamespacedPod(debugger.metadata.namespace).obj
        return debugger

    @staticmethod
    def exec_on_node(pod_name: str, node_name: str, cmd):
        node_runner = RobustaPod.create_debugger_pod(pod_name, node_name)
        try:
            node_runner.exec(f"nsenter -t 1 -a {cmd}")
        finally:
            node_runner.delete()

    @staticmethod
    def run_debugger_pod(node_name: str, pod_image: str, env: Optional[List[EnvVar]] = None,
                         mount_host_root: bool = False) -> str:
        debugger = RobustaPod.create_debugger_pod(node_name, node_name, pod_image, env=env, mount_host_root=mount_host_root)
        try:
            pod_name = debugger.metadata.name
            pod_namespace = debugger.metadata.namespace
            pod_status = wait_for_pod_status(pod_name, pod_namespace, SUCCEEDED_STATE, 360, 0.2)
            if pod_status != SUCCEEDED_STATE:
                raise Exception(f"pod {pod_name} in {pod_namespace} failed to complete. It is in state {pod_status}")

            return debugger.get_logs()
        finally:
            RobustaPod.deleteNamespacedPod(
                debugger.metadata.name, debugger.metadata.namespace
            )

    @staticmethod
    def exec_in_debugger_pod(
        pod_name: str, node_name: str, cmd, debug_image=PYTHON_DEBUGGER_IMAGE
    ) -> str:
        debugger = RobustaPod.create_debugger_pod(pod_name, node_name, debug_image)
        try:
            return debugger.exec(cmd)
        finally:
            RobustaPod.deleteNamespacedPod(
                debugger.metadata.name, debugger.metadata.namespace
            )

    def get_processes(self) -> List[Process]:
        output = RobustaPod.exec_in_debugger_pod(
            self.metadata.name,
            self.spec.nodeName,
            f"debug-toolkit pod-ps {self.metadata.uid}",
        )
        processes = ProcessList(**json.loads(output))
        return processes.processes

    def get_images(self) -> Dict[str, str]:
        return get_images(self.spec.containers)

    def has_direct_owner(self, owner_uid) -> bool:
        for owner in self.metadata.ownerReferences:
            if owner.uid == owner_uid:
                return True
        return False

    def has_toleration(self, toleration_key):
        return any(
            toleration_key == toleration.key for toleration in self.spec.tolerations
        )

    def has_cpu_limit(self) -> bool:
        for container in self.spec.containers:
            if container.resources and container.resources.limits.get("cpu"):
                return True
        return False

    def upload_file(self, path: str, contents: bytes, container: Optional[str] = None):
        if container is None:
            container = self.spec.containers[0].name
            logging.info(
                f"no container name given when uploading file, so choosing first container: {container}"
            )
        upload_file(
            self.metadata.name,
            path,
            contents,
            namespace=self.metadata.namespace,
            container=container,
        )

    @staticmethod
    def find_pods_with_direct_owner(
        namespace: str, owner_uid: str
    ) -> List["RobustaPod"]:
        all_pods: List["RobustaPod"] = PodList.listNamespacedPod(namespace).obj.items
        return list(filter(lambda p: p.has_direct_owner(owner_uid), all_pods))

    @staticmethod
    def find_pod(name_prefix, namespace) -> "RobustaPod":
        pods: PodList = PodList.listNamespacedPod(namespace).obj
        for pod in pods.items:
            if pod.metadata.name.startswith(name_prefix):
                # we serialize and then deserialize to work around https://github.com/haxsaw/hikaru/issues/15
                return hikaru.from_dict(pod.to_dict(), cls=RobustaPod)
        raise Exception(
            f"No pod exists in namespace '{namespace}' with name prefix '{name_prefix}'"
        )

    # TODO: replace with Hikaru Pod().read() but note that usage is slightly different as this is a staticmethod
    @staticmethod
    def read(name: str, namespace: str) -> "RobustaPod":
        """Read pod definition from the API server"""
        return Pod.readNamespacedPod(name, namespace).obj


class RobustaDeployment(Deployment):
    @classmethod
    def from_image(cls: Type[T], name, image="busybox", cmd=None) -> T:
        obj: RobustaDeployment = hikaru.from_dict(
            yaml.safe_load(get_deployment_yaml(name, image)), RobustaDeployment
        )
        obj.spec.template.spec.containers[0].command = prepare_pod_command(cmd)
        return obj

    def get_images(self) -> Dict[str, str]:
        return get_images(self.spec.template.spec.containers)


class RobustaJob(Job):
    def get_pods(self) -> List[RobustaPod]:
        """
        gets the pods associated with a job
        """
        pods: PodList = PodList.listNamespacedPod(
            self.metadata.namespace, label_selector=f"job-name = {self.metadata.name}"
        ).obj
        # we serialize and then deserialize to work around https://github.com/haxsaw/hikaru/issues/15
        return [hikaru.from_dict(pod.to_dict(), cls=RobustaPod) for pod in pods.items]

    def get_single_pod(self) -> RobustaPod:
        """
        like get_pods() but verifies that only one pod is associated with the job and returns that pod
        """
        pods = self.get_pods()
        if len(pods) != 1:
            raise Exception(f"got more pods than expected for job: {pods}")
        return pods[0]

    @classmethod
    def run_simple_job_spec(cls, spec, name, timeout) -> str:
        job = RobustaJob(
            metadata=ObjectMeta(
                namespace=INSTALLATION_NAMESPACE, name=to_kubernetes_name(name)
            ),
            spec=JobSpec(
                backoffLimit=0,
                template=PodTemplateSpec(
                    spec=spec,
                ),
            ),
        )
        try:
            job = job.createNamespacedJob(job.metadata.namespace).obj
            job = hikaru.from_dict(
                job.to_dict(), cls=RobustaJob
            )  # temporary workaround for hikaru bug #15
            job: RobustaJob = wait_until_job_complete(job, timeout)
            job = hikaru.from_dict(
                job.to_dict(), cls=RobustaJob
            )  # temporary workaround for hikaru bug #15
            pod = job.get_single_pod()
            return pod.get_logs()
        finally:
            job.deleteNamespacedJob(
                job.metadata.name,
                job.metadata.namespace,
                propagation_policy="Foreground",
            )

    @classmethod
    def run_simple_job(cls, image, command, timeout) -> str:
        spec = PodSpec(
            containers=[
                Container(
                    name=to_kubernetes_name(image),
                    image=image,
                    command=prepare_pod_command(command),
                )
            ],
            restartPolicy="Never",
        )
        return cls.run_simple_job_spec(spec, name=image, timeout=timeout)


hikaru.register_version_kind_class(RobustaPod, Pod.apiVersion, Pod.kind)
hikaru.register_version_kind_class(
    RobustaDeployment, Deployment.apiVersion, Deployment.kind
)
hikaru.register_version_kind_class(RobustaJob, Job.apiVersion, Job.kind)



from pydantic import BaseModel
from typing import Dict, Optional, Type, List

from ..model.events import ExecutionBaseEvent
from ..model.k8s_operation_type import K8sOperationType
from ..reporting.base import Finding
import abc


class TriggerEvent(BaseModel):
    @abc.abstractmethod
    def get_event_name(self) -> str:
        """Return trigger event name"""
        return ""


class BaseTrigger(BaseModel):
    def get_trigger_event(self) -> str:
        pass

    def should_fire(self, event: TriggerEvent, playbook_id: str):
        return True

    def build_execution_event(
        self, event: TriggerEvent, sink_findings: Dict[str,List[Finding]]
    ) -> Optional[ExecutionBaseEvent]:
        pass

    @staticmethod
    def get_execution_event_type() -> Type[ExecutionBaseEvent]:
        pass



from typing import List

from .msteams_base import MsTeamsBase


class MsTeamsContainer(MsTeamsBase):
    
    def __init__(self, key: str, elements: List[MsTeamsBase]):
        super().__init__(self.__container(key, elements))

    def __container(self, key: str, elements: List[MsTeamsBase]):
        block = {
            'type': "Container",
            "style": "accent",
            "isVisible": False,
            "bleed": False,
            "items": self.__get_items(elements),
        }
        if key is not None:
            block["id"] = key
        return block

    @classmethod
    def __get_items(cls, elements: List[MsTeamsBase]):
        return [elem.get_map_value() for elem in elements]



from ..sink_base_params import SinkBaseParams
from ..sink_config import SinkConfigBase


class MsTeamsSinkParams(SinkBaseParams):
    webhook_url: str


class MsTeamsSinkConfigWrapper(SinkConfigBase):
    ms_teams_sink: MsTeamsSinkParams

    def get_params(self) -> SinkBaseParams:
        return self.ms_teams_sink




import os
import os.path
from inspect import getmembers
import manhole

from .telemetry_service import TelemetryService, TelemetryLevel
from .log_init import logging, init_logging
from .web import Web
from ..core.playbooks.playbooks_event_handler_impl import PlaybooksEventHandlerImpl
from .. import api as robusta_api
from .config_loader import ConfigLoader
from ..model.config import Registry
from ..core.model.env_vars import ROBUSTA_TELEMETRY_ENDPOINT, SEND_ADDITIONAL_TELEMETRY, \
 ENABLE_TELEMETRY, TELEMETRY_PERIODIC_SEC

def main():
    init_logging()
    registry = Registry()
    event_handler = PlaybooksEventHandlerImpl(registry)
    loader = ConfigLoader(registry, event_handler)

    if ENABLE_TELEMETRY:    
        telemetry_service = TelemetryService(
            telemetry_level= TelemetryLevel.ERROR if SEND_ADDITIONAL_TELEMETRY else TelemetryLevel.USAGE,
            endpoint=ROBUSTA_TELEMETRY_ENDPOINT,
            periodic_time_sec= TELEMETRY_PERIODIC_SEC,
            registry= registry,
            )
    else:
        logging.info(f"Telemetry is disabled.")

    if os.environ.get("ENABLE_MANHOLE", "false").lower() == "true":
        manhole.install(locals=dict(getmembers(robusta_api)))

    
    Web.init(event_handler, loader)
    Web.run()  # blocking
    loader.close()


if __name__ == "__main__":
    main()



from enum import Enum

from hikaru.model import Pod
from pydantic import BaseModel

k8s_memory_factors = {
    "K": 1000,
    "M": 1000*1000,
    "G": 1000*1000*1000,
    "P": 1000*1000*1000*1000,
    "E": 1000*1000*1000*1000*1000,
    "Ki": 1024,
    "Mi": 1024*1024,
    "Gi": 1024*1024*1024,
    "Pi": 1024*1024*1024*1024,
    "Ei": 1024*1024*1024*1024*1024
}


class PodResources(BaseModel):
    pod_name: str
    cpu: float
    memory: int

    @staticmethod
    def parse_cpu(cpu: str) -> float:
        if not cpu:
            return 0.0
        if "m" in cpu:
            return round(float(cpu.replace("m", "").strip()) / 1000, 3)
        return round(float(cpu), 3)

    @staticmethod
    def parse_mem(mem: str) -> int:
        if not mem:
            return 0
        num_of_bytes = PodResources.get_number_of_bytes_from_kubernetes_mem_spec(mem)
        return int(num_of_bytes/(1024*1024))

    @staticmethod
    def get_number_of_bytes_from_kubernetes_mem_spec(mem_spec: str) -> int:
        if len(mem_spec) > 2 and mem_spec[-2:] in k8s_memory_factors:
            return int(mem_spec[:-2]) * k8s_memory_factors[mem_spec[-2:]]

        if len(mem_spec) > 1 and mem_spec[-1] in k8s_memory_factors:
            return int(mem_spec[:-1]) * k8s_memory_factors[mem_spec[-1]]

        raise Exception("number of bytes could not be extracted from memory spec: " + mem_spec)


def pod_restarts(pod: Pod) -> int:
    return sum([status.restartCount for status in pod.status.containerStatuses])


ResourceAttributes = Enum("ResourceAttributes", "requests limits")


def pod_requests(pod: Pod) -> PodResources:
    return pod_resources(pod, ResourceAttributes.requests)


def pod_limits(pod: Pod) -> PodResources:
    return pod_resources(pod, ResourceAttributes.limits)


def pod_resources(pod: Pod, resource_attribute: ResourceAttributes) -> PodResources:
    pod_cpu_req: float = 0.0
    pod_mem_req: int = 0
    for container in pod.spec.containers:
        try:
            requests = container.object_at_path(
                ["resources", resource_attribute.name]  # requests or limits
            )
            pod_cpu_req += PodResources.parse_cpu(
                requests.get("cpu", 0.0)
            )
            pod_mem_req += PodResources.parse_mem(
                requests.get("memory", "0Mi")
            )
        except Exception:
            pass  # no requests on container, object_at_path throws error

    return PodResources(
        pod_name=pod.metadata.name,
        cpu=pod_cpu_req,
        memory=pod_mem_req,
    )



import os
import subprocess
import time
from contextlib import contextmanager
from typing import Optional

import click_spinner
import toml
import typer
import requests
from dpath.util import get

PLAYBOOKS_DIR = "playbooks/"


def namespace_to_kubectl(namespace: Optional[str]):
    if namespace is None:
        return ""
    else:
        return f"-n {namespace}"


def exec_in_robusta_runner(
    cmd,
    namespace: Optional[str],
    tries=1,
    time_between_attempts=10,
    error_msg="error running cmd",
):
    exec_cmd = [
        "kubectl",
        "exec",
        "-it",
        "deployment/robusta-runner",
        "-c",
        "runner",
    ]
    if namespace is not None:
        exec_cmd += ["-n", namespace]

    exec_cmd += ["--", "bash", "-c", cmd]

    typer.echo(f"running cmd: {cmd}")

    for _ in range(tries - 1):
        try:
            return subprocess.check_call(exec_cmd)
        except Exception as e:
            typer.secho(f"error: {error_msg}", fg="red")
            time.sleep(time_between_attempts)
    return subprocess.check_call(cmd)


def download_file(url, local_path):
    with click_spinner.spinner():
        response = requests.get(url)
        response.raise_for_status()
    with open(local_path, "wb") as f:
        f.write(response.content)


def log_title(title, color=None):
    typer.echo("=" * 70)
    typer.secho(title, fg=color)
    typer.echo("=" * 70)


def replace_in_file(path, original, replacement):
    with open(path) as r:
        text = r.read()
        if original not in text:
            raise Exception(
                f"Cannot replace text {original} in file {path} because it was not found"
            )
        text = text.replace(original, replacement)
    with open(path, "w") as w:
        w.write(text)


@contextmanager
def fetch_runner_logs(namespace: Optional[str], all_logs=False):
    start = time.time()
    try:
        yield
    finally:
        log_title("Fetching logs...")
        try:
            if all_logs:
                subprocess.check_call(
                    f"kubectl logs {namespace_to_kubectl(namespace)} deployment/robusta-runner -c runner",
                    shell=True,
                )
            else:
                subprocess.check_call(
                    f"kubectl logs {namespace_to_kubectl(namespace)} deployment/robusta-runner -c runner --since={int(time.time() - start + 1)}s",
                    shell=True,
                )
        except:
            log_title("Cannot fetch logs. robusta-runner not found", color="red")
            return


def get_package_name(playbooks_dir: str) -> str:
    with open(os.path.join(playbooks_dir, "pyproject.toml"), "r") as pyproj_toml:
        data = pyproj_toml.read()
        parsed = toml.loads(data)
        return get(parsed, "tool/poetry/name", default="")



from hikaru.model import Pod, ObjectReference

from .base import FindingSubject
from .consts import FindingSubjectType


class KubeObjFindingSubject(FindingSubject):
    def __init__(
        self,
        obj=None,
        finding_subject_type: FindingSubjectType = None,
        should_add_node_name: bool = True,
    ):
        node_name = None
        if should_add_node_name:
            node_name = KubeObjFindingSubject.get_node_name(obj)
        if not finding_subject_type:
            finding_subject_type = FindingSubjectType.from_kind(obj.kind)
        super(KubeObjFindingSubject, self).__init__(
            obj.metadata.name, finding_subject_type, obj.metadata.namespace, node_name
        )

    @staticmethod
    def get_node_name(obj):
        if not obj:
            return None
        is_node_object = hasattr(obj, "kind") and obj.kind == "Node"
        if is_node_object:
            if isinstance(obj, ObjectReference):
                return obj.name

            return obj.metadata.name
        elif hasattr(obj, "spec") and hasattr(obj.spec, "nodeName"):
            return obj.spec.nodeName
        return None


class PodFindingSubject(FindingSubject):
    def __init__(self, pod: Pod = None):
        super(PodFindingSubject, self).__init__(
            pod.metadata.name,
            FindingSubjectType.TYPE_POD,
            pod.metadata.namespace,
            pod.spec.nodeName,
        )



import base64
import json
import logging
import time
import threading
import traceback
from collections import defaultdict

from hikaru.model import (
    Deployment,
    StatefulSetList,
    DaemonSetList,
    ReplicaSetList,
    Node,
    NodeList,
    Taint,
    NodeCondition,
    PodList,
    Pod,
)
from typing import List, Dict

from robusta.runner.web_api import WebApi
from .robusta_sink_params import RobustaSinkConfigWrapper, RobustaToken
from ...model.env_vars import DISCOVERY_PERIOD_SEC, PERIODIC_LONG_SEC
from ...model.nodes import NodeInfo
from ...model.pods import PodResources, pod_requests
from ...model.services import ServiceInfo
from ...reporting.base import Finding
from .dal.supabase_dal import SupabaseDal
from ..sink_base import SinkBase
from ...discovery.top_service_resolver import TopServiceResolver
from ...model.cluster_status import ClusterStatus


class RobustaSink(SinkBase):
    def __init__(self, sink_config: RobustaSinkConfigWrapper, registry):
        super().__init__(sink_config.robusta_sink, registry)
        self.token = sink_config.robusta_sink.token

        robusta_token = RobustaToken(**json.loads(base64.b64decode(self.token)))
        if self.account_id != robusta_token.account_id:
            logging.error(
                f"Account id configuration mismatch. "
                f"Global Config: {self.account_id} Robusta token: {robusta_token.account_id}."
                f"Using account id from Robusta token."
            )
        self.account_id = robusta_token.account_id

        self.dal = SupabaseDal(
            robusta_token.store_url,
            robusta_token.api_key,
            robusta_token.account_id,
            robusta_token.email,
            robusta_token.password,
            sink_config.robusta_sink.name,
            self.cluster_name,
            self.signing_key,
        )

        self.last_send_time = 0
        self.__update_cluster_status()  # send runner version initially, then force prometheus alert time periodically.

        # start cluster discovery
        self.__active = True
        self.__discovery_period_sec = DISCOVERY_PERIOD_SEC
        self.__services_cache: Dict[str, ServiceInfo] = {}
        self.__nodes_cache: Dict[str, NodeInfo] = {}
        self.__thread = threading.Thread(target=self.__discover_cluster)
        self.__thread.start()

    def __assert_node_cache_initialized(self):
        if not self.__nodes_cache:
            logging.info("Initializing nodes cache")
            for node in self.dal.get_active_nodes():
                self.__nodes_cache[node.name] = node

    def stop(self):
        self.__active = False

    def write_finding(self, finding: Finding, platform_enabled: bool):
        self.dal.persist_finding(finding)

    # service discovery impl
    def __publish_service(self, service_info: ServiceInfo):
        logging.debug(f"publishing to {self.sink_name} service {service_info} ")
        self.dal.persist_service(service_info)

    def __is_cached(self, service_info: ServiceInfo):
        cache_key = service_info.get_service_key()
        return self.__services_cache.get(cache_key) is not None

    def __publish_new_services(self, active_services: List):
        active_services_keys = set()
        for service in active_services:
            service_info = ServiceInfo(
                name=service.metadata.name,
                namespace=service.metadata.namespace,
                service_type=service.kind,
            )
            cache_key = service_info.get_service_key()
            active_services_keys.add(cache_key)
            cached_service = self.__services_cache.get(cache_key)
            if not cached_service or cached_service != service_info:
                self.__publish_service(service_info)
                self.__services_cache[cache_key] = service_info

        # delete cached services that aren't active anymore
        cache_keys = list(self.__services_cache.keys())
        for service_key in cache_keys:
            if service_key not in active_services_keys:
                del self.__services_cache[service_key]

        # handle delete services
        persisted_services = self.dal.get_active_services()
        deleted_services = [
            service for service in persisted_services if not self.__is_cached(service)
        ]
        for deleted_service in deleted_services:
            deleted_service.deleted = True
            self.__publish_service(deleted_service)

        # save the cached services in the resolver.
        TopServiceResolver.store_cached_services(list(self.__services_cache.values()))

    def __get_events_history(self):
        try:
            logging.info("Getting events history")
            # we will need the services in cache before the event history is run to guess service name
            self.__discover_services()
            response = WebApi.run_manual_action(
                action_name="event_history",
                sinks=[self.sink_name],
                retries=4,
                timeout_delay=30,
            )
            if response != 200:
                logging.error("Error running 'event_history'.")
            else:
                logging.info("Cluster historical data sent.")
        except Exception as e:
            logging.error(
                f"Error getting events history {e}\n" f"{traceback.format_exc()}"
            )

    def __discover_services(self):
        try:
            current_services = Deployment.listDeploymentForAllNamespaces().obj.items
            current_services.extend(
                StatefulSetList.listStatefulSetForAllNamespaces().obj.items
            )
            current_services.extend(
                DaemonSetList.listDaemonSetForAllNamespaces().obj.items
            )
            current_services.extend(
                [
                    rs
                    for rs in ReplicaSetList.listReplicaSetForAllNamespaces().obj.items
                    if rs.metadata.ownerReferences is None
                ]
            )
            self.__publish_new_services(current_services)
        except Exception as e:
            logging.error(
                f"Failed to run periodic service discovery for {self.sink_name}",
                exc_info=True,
            )

    @classmethod
    def __to_taint_str(cls, taint: Taint) -> str:
        return f"{taint.key}={taint.value}:{taint.effect}"

    @classmethod
    def __to_active_conditions_str(cls, conditions: List[NodeCondition]) -> str:
        if not conditions:
            return ""
        return ",".join(
            [
                f"{condition.type}:{condition.status}"
                for condition in conditions
                if condition.status != "False" or condition.type == "Ready"
            ]
        )

    @classmethod
    def __to_node_info(cls, node: Node) -> Dict:
        node_info = node.status.nodeInfo.to_dict() if node.status.nodeInfo else {}
        node_info["labels"] = node.metadata.labels
        node_info["annotations"] = node.metadata.annotations
        node_info["addresses"] = [addr.address for addr in node.status.addresses]
        return node_info

    @classmethod
    def __from_api_server_node(
        cls, api_server_node: Node, pod_requests_list: List[PodResources]
    ) -> NodeInfo:
        addresses = api_server_node.status.addresses
        external_addresses = [
            address for address in addresses if "externalip" in address.type.lower()
        ]
        external_ip = ",".join([addr.address for addr in external_addresses])
        internal_addresses = [
            address for address in addresses if "internalip" in address.type.lower()
        ]
        internal_ip = ",".join([addr.address for addr in internal_addresses])
        taints = ",".join(
            [cls.__to_taint_str(taint) for taint in api_server_node.spec.taints]
        )

        return NodeInfo(
            name=api_server_node.metadata.name,
            node_creation_time=api_server_node.metadata.creationTimestamp,
            internal_ip=internal_ip,
            external_ip=external_ip,
            taints=taints,
            conditions=cls.__to_active_conditions_str(
                api_server_node.status.conditions
            ),
            memory_capacity=PodResources.parse_mem(
                api_server_node.status.capacity.get("memory", "0Mi")
            ),
            memory_allocatable=PodResources.parse_mem(
                api_server_node.status.allocatable.get("memory", "0Mi")
            ),
            memory_allocated=sum([req.memory for req in pod_requests_list]),
            cpu_capacity=PodResources.parse_cpu(
                api_server_node.status.capacity.get("cpu", "0")
            ),
            cpu_allocatable=PodResources.parse_cpu(
                api_server_node.status.allocatable.get("cpu", "0")
            ),
            cpu_allocated=round(sum([req.cpu for req in pod_requests_list]), 3),
            pods_count=len(pod_requests_list),
            pods=",".join([pod_req.pod_name for pod_req in pod_requests_list]),
            node_info=cls.__to_node_info(api_server_node),
        )

    def __publish_new_nodes(
        self, current_nodes: NodeList, node_requests: Dict[str, List[PodResources]]
    ):
        # convert to map
        curr_nodes = {}
        for node in current_nodes.items:
            curr_nodes[node.metadata.name] = node

        # handle deleted nodes
        cache_keys = list(self.__nodes_cache.keys())
        for node_name in cache_keys:
            if not curr_nodes.get(node_name):  # node doesn't exist any more, delete it
                self.__nodes_cache[node_name].deleted = True
                self.dal.publish_node(self.__nodes_cache[node_name])
                del self.__nodes_cache[node_name]

        # new or changed nodes
        for node_name in curr_nodes.keys():
            updated_node = self.__from_api_server_node(
                curr_nodes.get(node_name), node_requests[node_name]
            )
            if (
                self.__nodes_cache.get(node_name) != updated_node
            ):  # node not in the cache, or changed
                self.dal.publish_node(updated_node)
                self.__nodes_cache[node_name] = updated_node

    def __discover_nodes(self):
        try:
            self.__assert_node_cache_initialized()
            current_nodes: NodeList = NodeList.listNode().obj
            node_requests = defaultdict(list)
            for status in ["Running", "Unknown", "Pending"]:
                pods: PodList = Pod.listPodForAllNamespaces(
                    field_selector=f"status.phase={status}"
                ).obj
                for pod in pods.items:
                    if pod.spec.nodeName:
                        node_requests[pod.spec.nodeName].append(pod_requests(pod))

            self.__publish_new_nodes(current_nodes, node_requests)
        except Exception as e:
            logging.error(
                f"Failed to run periodic nodes discovery for {self.sink_name}",
                exc_info=True,
            )

    def __update_cluster_status(self):
        try:
            cluster_status = ClusterStatus(
                cluster_id=self.cluster_name,
                version=self.registry.get_telemetry().runner_version,
                last_alert_at=self.registry.get_telemetry().last_alert_at,
                account_id=self.account_id,
            )

            self.dal.publish_cluster_status(cluster_status)
        except Exception as e:
            logging.error(
                f"Failed to run periodic update cluster status for {self.sink_name}",
                exc_info=True,
            )

    def __run_events_history_thread(self):
        try:
            # here to prevent a race between checking and writing findings from robusta sink
            if self.dal.has_cluster_findings():
                logging.info("Cluster already has historical data, No history pulled.")
                return
            thread = threading.Thread(target=self.__get_events_history)
            thread.start()
        except:
            logging.error(f"Failed to run events history thread")

    def __discover_cluster(self):
        logging.info("Cluster discovery initialized")
        self.__run_events_history_thread()
        while self.__active:
            self.__periodic_cluster_status()
            self.__discover_services()
            self.__discover_nodes()
            time.sleep(self.__discovery_period_sec)

        logging.info(f"Service discovery for sink {self.sink_name} ended.")

    def __periodic_cluster_status(self):
        if self.registry.get_telemetry().last_alert_at:
            if time.time() - self.last_send_time > PERIODIC_LONG_SEC:
                self.last_send_time = time.time()
                self.__update_cluster_status()



import logging
import hashlib
import hmac
from typing import Optional, List
from uuid import UUID

from pydantic import BaseModel
import requests

from ..model.env_vars import RELAY_EXTERNAL_ACTIONS_URL


class ActionRequestBody(BaseModel):
    account_id: str
    cluster_name: str
    action_name: str
    timestamp: int
    action_params: dict = None
    sinks: Optional[List[str]] = None
    origin: str = None


class ExternalActionRequest(BaseModel):
    body: ActionRequestBody
    signature: str = ""         # Used for signature based auth protocol option
    partial_auth_a: str = ""    # Auth for public key auth protocol option - should be added by the client
    partial_auth_b: str = ""    # Auth for public key auth protocol option - should be added by the relay
    request_id: str = ""        # If specified, should return a sync response using the specified request_id
    no_sinks: bool = False      # Indicates not to send to sinks at all. The request body has a sink list,
                                # however an empty sink list means using the server default sinks


class PartialAuth(BaseModel):
    hash: str
    key: UUID


def sign_action_request(body: BaseModel, signing_key: str):
    format_req = str.encode(f"v0:{body.json(exclude_none=True, sort_keys=True)}")
    if not signing_key:
        raise Exception("Signing key not available. Cannot sign action request")
    request_hash = hmac.new(
        signing_key.encode(), format_req, hashlib.sha256
    ).hexdigest()
    return f"v0={request_hash}"


class OutgoingActionRequest:
    @staticmethod
    def send(body: ActionRequestBody, signing_key: str):
        signature = sign_action_request(body, signing_key)
        action_request = ExternalActionRequest(
            signature=signature,
            body=body,
        )

        resp = requests.post(RELAY_EXTERNAL_ACTIONS_URL, json=action_request.dict())
        if resp.status_code != 200:
            logging.error(
                f"Failed to send external action request {resp.status_code} {resp.reason}"
            )



import textwrap

import requests
from typing import List

from .webhook_sink_params import WebhookSinkConfigWrapper
from ..transformer import Transformer
from ...reporting import HeaderBlock, ListBlock, JsonBlock, KubernetesDiffBlock, MarkdownBlock
from ...reporting.base import Finding, BaseBlock
from ..sink_base import SinkBase


class WebhookSink(SinkBase):
    def __init__(self, sink_config: WebhookSinkConfigWrapper, registry):
        super().__init__(sink_config.webhook_sink, registry)

        self.url = sink_config.webhook_sink.url
        self.size_limit = sink_config.webhook_sink.size_limit

    def write_finding(self, finding: Finding, platform_enabled: bool):
        message_lines: List[str] = [finding.title]
        if platform_enabled:
            message_lines.append(f"Investigate: {finding.investigate_uri}")
        message_lines.append(f"Source: {self.cluster_name}")
        message_lines.append(finding.description)

        message = ""

        for enrichment in finding.enrichments:
            for block in enrichment.blocks:
                message_lines.extend(self.__to_unformatted_text(block))

        for line in [line for line in message_lines if line]:
            wrapped = textwrap.dedent(
                f"""
                {line}
                """
            )
            if len(message) + len(wrapped) >= self.size_limit:
                break
            message += wrapped

        requests.post(self.url, data=message)

    @classmethod
    def __to_clear_text(cls, markdown_text: str) -> str:
        # just create a readable links format
        links = Transformer.get_markdown_links(markdown_text)
        for link in links:
            # take only the data between the first '<' and last '>'
            splits = link[1:-1].split("|")
            if len(splits) == 2:  # don't replace unexpected strings
                replacement = f"{splits[1]}: {splits[0]}"
                markdown_text = markdown_text.replace(link, replacement)

        return markdown_text

    def __to_unformatted_text(cls, block: BaseBlock) -> List[str]:
        lines = []
        if isinstance(block, HeaderBlock):
            lines.append(block.text)
        elif isinstance(block, ListBlock):
            lines.extend([cls.__to_clear_text(item) for item in block.items])
        elif isinstance(block, MarkdownBlock):
            lines.append(cls.__to_clear_text(block.text))
        elif isinstance(block, JsonBlock):
            lines.append(block.json_str)
        elif isinstance(block, KubernetesDiffBlock):
            for diff in block.diffs:
                lines.append(f"*{'.'.join(diff.path)}*: {diff.other_value} ==> {diff.value}")
        return lines



from ..sink_config import SinkConfigBase
from ..sink_base_params import SinkBaseParams


class SlackSinkParams(SinkBaseParams):
    slack_channel: str
    api_key: str


class SlackSinkConfigWrapper(SinkConfigBase):
    slack_sink: SlackSinkParams

    def get_params(self) -> SinkBaseParams:
        return self.slack_sink



import hashlib
import logging
import os
import shutil
import subprocess
import textwrap
import threading
from collections import defaultdict, namedtuple
import traceback
import re
from typing import List, Dict

from ...core.model.env_vars import GIT_MAX_RETRIES

GIT_DIR_NAME = "robusta-git"
REPO_LOCAL_BASE_DIR = os.path.join(
    os.environ.get("REPO_LOCAL_BASE_DIR", "/app"), GIT_DIR_NAME
)
SSH_ROOT_DIR = os.environ.get("SSH_ROOT_DIR", "/root/.ssh")

GIT_SSH_PREFIX = "git@"
GIT_HTTPS_PREFIX = "https://git"
LOCAL_PATH_URL_PREFIX = "file://"


class GitRepoManager:

    manager_lock = threading.Lock()
    repo_map = defaultdict(None)

    @staticmethod
    def get_git_repo(git_repo_url: str, git_key: str):
        with GitRepoManager.manager_lock:
            repo = GitRepoManager.repo_map.get(git_repo_url)
            if repo is not None:
                return repo
            repo = GitRepo(git_repo_url, git_key)
            GitRepoManager.repo_map[git_repo_url] = repo
            return repo

    @staticmethod
    def remove_git_repo(git_repo_url):
        with GitRepoManager.manager_lock:
            del GitRepoManager.repo_map[git_repo_url]

    @staticmethod
    def clear_git_repos():
        with GitRepoManager.manager_lock:
            GitRepoManager.repo_map.clear()


SingleChange = namedtuple("SingleChange", "commit_date commit_message")
ClusterChanges = Dict[str, List[SingleChange]]


class GitRepo:

    initialized: bool = False

    def __init__(self, git_repo_url: str, git_key: str):
        GitRepo.init()
        self.git_repo_url = git_repo_url
        self.env = os.environ.copy()
        ssh_key_option = ""
        if git_key:  # Add ssh key for non-public repositories
            key_file_name = self.init_key(git_key)
            ssh_key_option = f"-i {key_file_name}"

        self.env["GIT_SSH_COMMAND"] = f"ssh {ssh_key_option} -o IdentitiesOnly=yes"
        self.repo_lock = threading.RLock()
        self.repo_name = os.path.splitext(os.path.basename(git_repo_url))[0]
        self.repo_local_path = os.path.join(REPO_LOCAL_BASE_DIR, self.repo_name)
        self.init_repo()

    def init_key(self, git_key):
        url_hash = hashlib.sha1(self.git_repo_url.encode("utf-8")).hexdigest()
        key_file_name = os.path.join(REPO_LOCAL_BASE_DIR, url_hash)
        if os.path.exists(key_file_name):
            return key_file_name

        # Secret managers might remove last empty line on private key, causing it to be illegal
        if not git_key.endswith("\n"):
            git_key = git_key + "\n"

        with open(key_file_name, "w") as key_file:
            key_file.write(textwrap.dedent(f"{git_key}"))
        os.chmod(key_file_name, 0o400)
        if not os.path.exists(SSH_ROOT_DIR):
            os.mkdir(SSH_ROOT_DIR)
        os.system(f"ssh-keyscan -H github.com bitbucket.org >> {SSH_ROOT_DIR}/known_hosts")
        return key_file_name

    @staticmethod
    def init():
        if GitRepo.initialized:
            return
        try:
            os.makedirs(REPO_LOCAL_BASE_DIR, exist_ok=True)
        except Exception as e:
            logging.error(
                f"Failed to create git audit base path {REPO_LOCAL_BASE_DIR}",
                exc_info=True,
            )
            raise e
        GitRepo.initialized = True

    def __exec_git_cmd(self, cmd: List[str]):
        shell = False
        if os.name == "nt":
            shell = True

        result = subprocess.run(
            cmd,
            cwd=self.repo_local_path,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=shell,
            env=self.env,
        )
        if result.returncode:
            logging.error(
                f"running command {cmd} failed with returncode={result.returncode}"
            )
            logging.error(f"stdout={result.stdout.decode()}")
            logging.error(f"stderr={result.stderr.decode()}")
            raise Exception(f"Error running git command: {cmd}")

        return result.stdout.decode()

    def init_repo(self):
        with self.repo_lock:
            if os.path.exists(self.repo_local_path):
                logging.info(f"Deleting local repo before init {self.repo_local_path}")
                shutil.rmtree(self.repo_local_path)

            logging.info(
                f"Cloning git repo {self.git_repo_url}. repo name {self.repo_name}"
            )
            os.makedirs(self.repo_local_path, exist_ok=True)
            self.__exec_git_cmd(
                ["git", "clone", self.git_repo_url, self.repo_local_path]
            )
            self.__exec_git_cmd(
                ["git", "config", "--global", "user.email", "runner@robusta.dev"]
            )
            self.__exec_git_cmd(
                ["git", "config", "--global", "user.name", "Robusta Runner"]
            )

    def commit(
        self,
        file_data: str,
        file_path: str,
        file_name,
        commit_message,
        cluster_name: str,
    ):
        with self.repo_lock:
            file_local_path = os.path.join(self.repo_local_path, file_path)
            try:
                os.makedirs(file_local_path, exist_ok=True)
                git_file_name = os.path.join(file_local_path, file_name)
                with open(git_file_name, "w") as git_file:
                    git_file.write(file_data)

                self.__exec_git_cmd(["git", "add", git_file_name])
                self.__exec_git_cmd(
                    [
                        "git",
                        "commit",
                        "-m",
                        self.__cluster_commit_msg(commit_message, cluster_name),
                        "--allow-empty",
                    ]
                )
            except Exception as e:
                logging.error(
                    f"Commit file failed {self.repo_local_path} {file_path} {file_name}",
                    exc_info=True,
                )
                GitRepoManager.remove_git_repo(self.git_repo_url)
                raise e

    @classmethod
    def __cluster_commit_msg(cls, msg: str, cluster_name: str):
        return f"Cluster {cluster_name}::{msg}"

    def push(self):
        with self.repo_lock:
            max_retries = GIT_MAX_RETRIES
            while max_retries > 0:
                try:
                    self.__exec_git_cmd(["git", "push"])
                    return
                except Exception as e:
                    max_retries -= 1
                    if max_retries > 0:
                        self.pull_rebase()
                    else:
                        GitRepoManager.remove_git_repo(self.git_repo_url)
                        logging.error(
                            f"Push failed {self.repo_local_path}", exc_info=True
                        )
                        raise e

    def pull_rebase(self):
        with self.repo_lock:
            self.__exec_git_cmd(["git", "pull", "--rebase", "-Xtheirs"])

    def cluster_changes(self, since_minutes: int = 20) -> ClusterChanges:
        cluster_changes = defaultdict(list)
        with self.repo_lock:
            self.pull_rebase()
            log = self.__exec_git_cmd(
                ["git", "log", f"--since='{since_minutes} minutes'"]
            )
            commit_date = ""
            for line in log.split("\n"):
                line = line.strip()
                if not line or line.startswith("Author") or line.startswith("commit"):
                    continue
                elif line.startswith("Date"):
                    commit_date = line.replace("Date:", "").strip()
                else:  # this is the commit message
                    if line.startswith("Cluster "):
                        line_suffix = re.sub("Cluster ", "", line)
                        cluster = re.sub("::.*", "", line_suffix)
                        commit_message = re.sub(".*::", "", line_suffix)
                    else:
                        cluster = "Unknown"
                        commit_message = line
                    cluster_changes[cluster].append(
                        SingleChange(commit_date, commit_message)
                    )

            return cluster_changes

    def commit_push(
        self,
        file_data: str,
        file_path: str,
        file_name,
        commit_message: str,
        cluster_name: str,
    ):
        with self.repo_lock:
            self.commit(file_data, file_path, file_name, commit_message, cluster_name)
            self.push()

    def delete(self, file_path: str, file_name, commit_message: str, cluster_name: str):
        with self.repo_lock:
            file_local_path = os.path.join(self.repo_local_path, file_path)
            if not os.path.exists(
                file_local_path
            ):  # file not in repo. Might have been added before the audit playbook was configured
                return

            try:
                os.remove(os.path.join(file_local_path, file_name))
                self.__exec_git_cmd(
                    [
                        "git",
                        "commit",
                        "-m",
                        self.__cluster_commit_msg(commit_message, cluster_name),
                    ]
                )
            except Exception as e:
                logging.error(
                    f"Commit file failed {self.repo_local_path} {file_path} {file_name}",
                    exc_info=True,
                )
                GitRepoManager.remove_git_repo(self.git_repo_url)
                raise e

    def delete_push(
        self, file_path: str, file_name, commit_message: str, cluster_name: str
    ):
        with self.repo_lock:
            self.delete(file_path, file_name, commit_message, cluster_name)
            self.push()



from pydantic import BaseModel, PrivateAttr, validator
from typing import Optional

from robusta.core.playbooks.playbook_utils import replace_env_vars_values


class PlaybookAction(BaseModel):
    action_name: str
    action_params: Optional[dict]
    _func_hash: str = PrivateAttr()

    def set_func_hash(self, func_hash):
        self._func_hash = func_hash

    def as_str(self):
        return self._func_hash + self.json()

    @validator('action_params')
    def env_var_params(cls, action_params: Optional[dict]):
        if action_params:
            action_params = replace_env_vars_values(action_params)
        return action_params



from typing import List, Optional
from pydantic import SecretStr
from ...utils.documented_pydantic import DocumentedModel
from enum import Enum, auto


class ChartValuesFormat(Enum):
    """
    Format option for chart rendering
    """
    Plain = auto()
    Bytes = auto()
    Percentage = auto()


class ResourceChartItemType(Enum):
    """
    Item selection for Alert resource enricher
    """
    Pod = auto()
    Node = auto()


class ResourceChartResourceType(Enum):
    """
    Resource selection for resource enricher(s)
    """
    CPU = auto()
    Memory = auto()
    Disk = auto()


class ActionParams(DocumentedModel):
    """
    Base class for all Action parameter classes.
    """

    pass


class RateLimitParams(ActionParams):
    """
    :var rate_limit: Rate limit the execution of this action (Seconds).
    """

    rate_limit: int = 3600


class FindingKeyParams(ActionParams):
    """
    :var finding_key: Specify the finding identifier, to reference it in other actions.
    """

    finding_key: str = "DEFAULT"


class BashParams(ActionParams):
    """
    :var bash_command: Bash command to execute on the target.

    :example bash_command: ls -l /etc/data/db
    """

    bash_command: str


class PrometheusParams(ActionParams):
    """
    :var prometheus_url: Prometheus url. If omitted, we will try to find a prometheus instance in the same cluster

    :example prometheus_url: "http://prometheus-k8s.monitoring.svc.cluster.local:9090"
    """

    prometheus_url: str = None


class CustomGraphEnricherParams(PrometheusParams):
    """
    :var promql_query: Promql query. You can use $pod, $node and $node_internal_ip to template (see example). For more information, see https://prometheus.io/docs/prometheus/latest/querying/basics/
    :var graph_title: A nicer name for the Prometheus query.
    :var graph_duration_minutes: Graph duration is minutes.
    :var chart_values_format: Customize the y-axis labels with one of: Plain, Bytes, Percentage (see ChartValuesFormat)

    :example promql_query: instance:node_cpu_utilisation:rate5m{job="node-exporter", instance=~"$node_internal_ip:[0-9]+", cluster=""} != 0
    :example graph_title: CPU Usage for this nod
    """

    promql_query: str
    graph_title: Optional[str] = None
    graph_duration_minutes: int = 60
    chart_values_format: str = 'Plain'


class ResourceGraphEnricherParams(PrometheusParams):
    """
    :var resource_type: one of: CPU, Memory, Disk (see ResourceChartResourceType)
    :var graph_duration_minutes: Graph duration is minutes. Default is 60.

    :example resource_type: Memory
    """
    resource_type: str
    graph_duration_minutes: int = 60


class AlertResourceGraphEnricherParams(ResourceGraphEnricherParams):
    """
    :var item_type: one of: Pod, Node (see ResourceChartItemType)
    :example item_type: Pod
    """
    item_type: str


class GrafanaParams(ActionParams):
    """
    :var grafana_url: http(s) url of grafana or None for autodetection of an in-cluster grafana
    :var grafana_api_key: grafana key with write permissions.
    :var grafana_dashboard_uid: dashboard ID as it appears in the dashboard's url

    :example grafana_url: http://grafana.namespace.svc
    :example grafana_dashboard_uid: 09ec8aa1e996d6ffcd6817bbaff4db1b
    """

    grafana_api_key: SecretStr
    grafana_dashboard_uid: str
    grafana_url: str = None


class GrafanaAnnotationParams(GrafanaParams):
    """
    :var grafana_dashboard_panel: when present, annotations will be added only to panels with this text in their title.
    :var cluster_name: written as one of the annotation's tags
    :var custom_tags: custom tags to add to the annotation
    """

    grafana_dashboard_panel: str = None
    cluster_name: str = None
    cluster_zone: str = None
    custom_tags: List[str] = None


class ProcessParams(ActionParams):
    """
    :var process_substring: process name (or substring).
    :var pid: pid
    :var interactive: if more than one process matches, interactively ask which process to choose.
    """

    process_substring: str = ""
    pid: int = None
    interactive: bool = True



import datetime
from slack_sdk import WebClient
from typing import Optional, List
from pydantic import BaseModel

import urllib.request

REMOTE_FEEDBACK_MESSAGE_ADDRESS = "https://docs.robusta.dev/extra/feedback_messages.json"


class SlackFeedbackMessage(BaseModel):
    minutes_from_now: int
    title: str
    other_sections: List[str]


class SlackFeedbackConfig(BaseModel):
    heads_up_message: str
    messages: List[SlackFeedbackMessage]


class SlackFeedbackMessagesSender(object):
    def __init__(self, slack_api_key: str, channel_name: str, account_id: str, debug: bool):
        self.slack_api_key = slack_api_key
        self.channel_name = channel_name
        self.account_id = account_id
        self.debug = debug

    def schedule_feedback_messages(self) -> Optional[str]:
        raw_feedback_messages = self._get_feedback_config_from_remote()
        slack_feedback_config: SlackFeedbackConfig = SlackFeedbackConfig.parse_raw(raw_feedback_messages)
        if len(slack_feedback_config.messages) == 0:
            return None
        for feedback_message in slack_feedback_config.messages:
            self._schedule_message(
                feedback_message.minutes_from_now,
                self._replace_account_id(feedback_message.title),
                list(map(self._replace_account_id, feedback_message.other_sections)))
        return slack_feedback_config.heads_up_message

    def _replace_account_id(self, text: str):
        return text.replace('$ACCOUNT_ID', self.account_id)

    @staticmethod
    def _get_feedback_config_from_remote() -> str:
        with urllib.request.urlopen(REMOTE_FEEDBACK_MESSAGE_ADDRESS) as response:
            return response.read()

    def _schedule_message(
            self,
            minutes_from_now: int,
            title: str,
            other_sections: List[str]
    ):
        slack_client = WebClient(token=self.slack_api_key)

        schedule_datetime = datetime.datetime.now() + datetime.timedelta(minutes=minutes_from_now)
        schedule_timestamp = schedule_datetime.strftime('%s')

        slack_client.chat_scheduleMessage(
            channel=self.channel_name,
            post_at=schedule_timestamp,
            text="Your feedback is important",
            blocks=self._gen_robusta_slack_message(title, other_sections),
            display_as_bot=True,
            unfurl_links=True,
            unfurl_media=True
        )

    @staticmethod
    def _gen_robusta_slack_message(title: str, other_sections: List[str]):
        blocks = [
            {
                "type": "header",
                "text": {
                    "type": "plain_text",
                    "text": title,
                    "emoji": True,
                },
            },
        ]
        additional_blocks = [{
            "type": "section",
            "text": {
                "type": "mrkdwn",
                "text": message
            },
        } for message in other_sections]
        blocks.extend(additional_blocks)
        return blocks



import logging
import hikaru
from typing import Dict, Any, Optional, List
from pydantic import BaseModel, PrivateAttr

from .autogenerated.events import KIND_TO_EVENT_CLASS
from .autogenerated.models import get_api_version
from .model_not_found_exception import ModelNotFoundException
from ..helper import prefix_match, exact_match
from ...core.playbooks.base_trigger import BaseTrigger
from ...core.playbooks.base_trigger import TriggerEvent
from ...core.model.k8s_operation_type import K8sOperationType
from ...core.model.events import ExecutionBaseEvent
from ...core.reporting.base import Finding


class IncomingK8sEventPayload(BaseModel):
    """
    The format of incoming payloads containing kubernetes events. This is mostly used for deserialization.
    """

    operation: str
    kind: str
    apiVersion: str = ""
    clusterUid: str
    description: str
    obj: Dict[Any, Any]
    oldObj: Optional[Dict[Any, Any]]


class K8sTriggerEvent(TriggerEvent):
    k8s_payload: IncomingK8sEventPayload

    def get_event_name(self) -> str:
        return K8sTriggerEvent.__name__


class K8sBaseTrigger(BaseTrigger):
    kind: str
    operation: K8sOperationType = None
    name_prefix: str = None
    namespace_prefix: str = None
    labels_selector: str = None
    _labels_map: Dict = PrivateAttr()

    def __init__(self, *args, **data):
        super().__init__(*args, **data)
        if self.labels_selector:
            labels = self.labels_selector.split(",")
            self._labels_map = {}
            for label in labels:
                label_parts = label.split("=")
                if len(label_parts) != 2:
                    msg = f"Illegal label selector {label}"
                    logging.error(msg)
                    raise Exception(msg)

                self._labels_map[label_parts[0].strip()] = label_parts[1].strip()

    def get_trigger_event(self):
        return K8sTriggerEvent.__name__

    def should_fire(self, event: TriggerEvent, playbook_id: str):
        if not isinstance(event, K8sTriggerEvent):
            return False

        k8s_trigger_event = K8sTriggerEvent(**event.dict())
        k8s_payload = k8s_trigger_event.k8s_payload
        if self.kind != "Any" and self.kind != k8s_payload.kind:
            return False

        if not exact_match(self.operation, K8sOperationType(k8s_payload.operation)):
            return False

        meta = k8s_payload.obj.get("metadata", {})
        if not prefix_match(self.name_prefix, meta.get("name", "")):
            return False

        if not prefix_match(self.namespace_prefix, meta.get("namespace", "")):
            return False

        labels_map = getattr(self, "_labels_map", None)
        if labels_map:
            obj_labels = meta.get("labels", {})
            for label_key, label_value in labels_map.items():
                if label_value != obj_labels.get(label_key, ""):
                    return False

        return True

    @classmethod
    def __parse_kubernetes_objs(cls, k8s_payload: IncomingK8sEventPayload):
        model_class = get_api_version(k8s_payload.apiVersion).get(k8s_payload.kind)
        if model_class is None:
            msg = f"classes for kind {k8s_payload.kind} cannot be found. skipping. description {k8s_payload.description}"
            logging.error(msg)
            raise ModelNotFoundException(msg)

        obj = hikaru.from_dict(k8s_payload.obj, cls=model_class)
        old_obj = None
        if k8s_payload.oldObj is not None:
            old_obj = hikaru.from_dict(k8s_payload.oldObj, cls=model_class)
        return obj, old_obj

    def build_execution_event(
        self, event: K8sTriggerEvent, sink_findings: Dict[str,List[Finding]]
    ) -> Optional[ExecutionBaseEvent]:
        # we can't use self.get_execution_event_type() because for KubernetesAnyAllChangesTrigger we need to filter out
        # stuff like ConfigMaps and we do that filtering here by checking if there is a real event_class
        # it might be better to move that filtering logic to should_fire() where it belongs and to use
        # self.get_execution_event_type() here instead of KIND_TO_EVENT_CLASS. Using KIND_TO_EVENT_CLASS leads to
        # inconsistencies for KubernetesAnyAllChangesTrigger (and possibly elsewhere) which claims in
        # get_execution_event_type() that it creates a KubernetesAnyChangeEvent object but it really creates
        # a different concrete event class using the logic below
        event_class = KIND_TO_EVENT_CLASS.get(event.k8s_payload.kind.lower())
        if event_class is None:
            logging.info(
                f"classes for kind {event.k8s_payload.kind} cannot be found. skipping. description {event.k8s_payload.description}"
            )
            return None
        (obj, old_obj) = self.__parse_kubernetes_objs(event.k8s_payload)
        operation_type = K8sOperationType(event.k8s_payload.operation)
        return event_class(
            sink_findings=sink_findings,
            operation=operation_type,
            description=event.k8s_payload.description.replace("\n", ""),
            obj=obj,
            old_obj=old_obj,
        )



import json
import logging
from typing import Union, Any


def load_json(s: Union[str, bytes]) -> Any:
    """
    Like json.loads() but prints the input if parsing fails
    """
    try:
        return json.loads(s)
    except json.decoder.JSONDecodeError:
        sep = "----------------"
        logging.error(f"could not parse json:\n{sep}\n{s}\n{sep}\n")
        raise



import re
import urllib.parse

import markdown2
from typing import List

from ...core.reporting import MarkdownBlock, BaseBlock, DividerBlock, JsonBlock, KubernetesDiffBlock, HeaderBlock, \
    ListBlock, TableBlock, tabulate


class Transformer:

    @staticmethod
    def get_markdown_links(markdown_data: str) -> List[str]:
        regex = "<.*?\\|.*?>"
        matches = re.findall(regex, markdown_data)
        links = []
        if matches:
            links = [
                match for match in matches if len(match) > 1
            ]  # filter out illegal matches
        return links

    @staticmethod
    def to_github_markdown(markdown_data: str, add_angular_brackets: bool = True) -> str:
        """Transform all occurrences of slack markdown, <URL|LINK TEXT>, to github markdown [LINK TEXT](URL)."""
        # some markdown parsers doesn't support angular brackets on links
        OPENING_ANGULAR = "<" if add_angular_brackets else ""
        CLOSING_ANGULAR = ">" if add_angular_brackets else ""
        matches = Transformer.get_markdown_links(markdown_data)
        for match in matches:
            # take only the data between the first '<' and last '>'
            splits = match[1:-1].split("|")
            if len(splits) == 2:  # don't replace unexpected strings
                parsed_url = urllib.parse.urlparse(splits[0])
                parsed_url = parsed_url._replace(path=urllib.parse.quote_plus(parsed_url.path, safe="/"))
                replacement = f"[{splits[1]}]({OPENING_ANGULAR}{parsed_url.geturl()}{CLOSING_ANGULAR})"
                markdown_data = markdown_data.replace(match, replacement)
        return re.sub(r"\*([^\*]*)\*", r"**\1**", markdown_data)

    @classmethod
    def __markdown_to_html(cls, mrkdwn_text: str) -> str:
        # replace links: from <http://url|name> to <a href="url">name</a>
        mrkdwn_links = re.findall(r"<[^\\|]*\|[^\>]*>", mrkdwn_text)
        for link in mrkdwn_links:
            link_content = link[1:-1]
            link_parts = link_content.split("|")
            mrkdwn_text = mrkdwn_text.replace(link, f"<a href=\"{link_parts[0]}\">{link_parts[1]}</a>")

        # replace slack markdown bold: from *bold text* to <b>bold text<b>  (markdown2 converts this to italic)
        mrkdwn_text = re.sub(r"\*([^\*]*)\*", r"<b>\1</b>", mrkdwn_text)

        # Note - markdown2 should be used after slack links already converted, otherwise it's getting corrupted!
        # Convert other markdown content
        return markdown2.markdown(mrkdwn_text)

    @classmethod
    def to_html(cls, blocks: List[BaseBlock]) -> str:
        lines = []
        for block in blocks:
            if isinstance(block, MarkdownBlock):
                if not block.text:
                    continue
                lines.append(f"{cls.__markdown_to_html(block.text)}")
            elif isinstance(block, DividerBlock):
                lines.append("-------------------")
            elif isinstance(block, JsonBlock):
                lines.append(block.json_str)
            elif isinstance(block, KubernetesDiffBlock):
                for diff in block.diffs:
                    lines.append(
                        cls.__markdown_to_html(f"*{'.'.join(diff.path)}*: {diff.other_value} ==> {diff.value}")
                    )
            elif isinstance(block, HeaderBlock):
                lines.append(f"<strong>{block.text}</strong>")
            elif isinstance(block, ListBlock):
                lines.extend(cls.__markdown_to_html(block.to_markdown().text))
            elif isinstance(block, TableBlock):
                if block.table_name:
                    lines.append(cls.__markdown_to_html(block.table_name))
                lines.append(
                    tabulate(block.render_rows(), headers=block.headers, tablefmt="html").replace("\n", "")
                )
        return "\n".join(lines)

    @classmethod
    def to_standard_markdown(cls, blocks: List[BaseBlock]) -> str:
        lines = []
        for block in blocks:
            if isinstance(block, MarkdownBlock):
                if not block.text:
                    continue
                lines.append(f"{cls.to_github_markdown(block.text, False)}")
            elif isinstance(block, DividerBlock):
                lines.append("-------------------")
            elif isinstance(block, JsonBlock):
                lines.append(block.json_str)
            elif isinstance(block, KubernetesDiffBlock):
                for diff in block.diffs:
                    lines.append(
                        f"**{'.'.join(diff.path)}**: {diff.other_value} ==> {diff.value}"
                    )
            elif isinstance(block, HeaderBlock):
                lines.append(f"**{block.text}**")
            elif isinstance(block, ListBlock):
                lines.extend(cls.to_github_markdown(block.to_markdown().text, False))
            elif isinstance(block, TableBlock):
                if block.table_name:
                    lines.append(cls.to_github_markdown(block.table_name, False))
                rendered_rows = block.render_rows()
                lines.append(
                    tabulate(rendered_rows, headers=block.headers, tablefmt="presto")
                )
        return "\n".join(lines)



from ...reporting import Finding
from ....integrations.msteams.sender import MsTeamsSender
from ..sink_base import SinkBase
from .msteams_sink_params import MsTeamsSinkConfigWrapper


class MsTeamsSink(SinkBase):
    def __init__(
        self,
        sink_config: MsTeamsSinkConfigWrapper,
        registry
    ):
        super().__init__(sink_config.ms_teams_sink, registry)
        self.webhook_url = sink_config.ms_teams_sink.webhook_url

    def write_finding(self, finding: Finding, platform_enabled: bool):
        MsTeamsSender.send_finding_to_ms_teams(self.webhook_url, finding, platform_enabled, self.cluster_name)



import base64
import json
import uuid

from typing import Dict, Any

from ...transformer import Transformer
from ....reporting.callbacks import ExternalActionRequestBuilder
from .....core.reporting import Finding, Enrichment, MarkdownBlock, logging, CallbackBlock, KubernetesDiffBlock, \
    HeaderBlock, ListBlock, TableBlock, FileBlock, DividerBlock


class ModelConversion:

    @staticmethod
    def to_finding_json(account_id: str, cluster_id: str, finding: Finding):
        finding_json = {
            "id": str(finding.id),
            "title": finding.title,
            "description": finding.description,
            "source": finding.source.value,
            "aggregation_key": finding.aggregation_key,
            "failure": finding.failure,
            "finding_type": finding.finding_type.value,
            "category": finding.category,
            "priority": finding.severity.name,
            "subject_type": finding.subject.subject_type.value,
            "subject_name": finding.subject.name,
            "subject_namespace": finding.subject.namespace,
            "subject_node": finding.subject.node,
            "service_key": finding.service_key,
            "cluster": cluster_id,
            "account_id": account_id,
        }

        if finding.creation_date:
            finding_json["creation_date"] = finding.creation_date

        return finding_json

    @staticmethod
    def to_evidence_json(
            account_id: str,
            cluster_id: str,
            sink_name: str,
            signing_key: str,
            finding_id: uuid,
            enrichment: Enrichment
        ) -> Dict[Any, Any]:
        structured_data = []
        for block in enrichment.blocks:
            if isinstance(block, MarkdownBlock):
                if not block.text:
                    continue
                structured_data.append(
                    {
                        "type": "markdown",
                        "data": Transformer.to_github_markdown(block.text),
                    }
                )
            elif isinstance(block, DividerBlock):
                structured_data.append({"type": "divider"})
            elif isinstance(block, FileBlock):
                last_dot_idx = block.filename.rindex(".")
                structured_data.append(
                    {
                        "type": block.filename[last_dot_idx + 1 :],
                        "data": str(base64.b64encode(block.contents)),
                    }
                )
            elif isinstance(block, HeaderBlock):
                structured_data.append({"type": "header", "data": block.text})
            elif isinstance(block, ListBlock):
                structured_data.append({"type": "list", "data": block.items})
            elif isinstance(block, TableBlock):
                if block.table_name:
                    structured_data.append({
                            "type": "markdown",
                            "data": Transformer.to_github_markdown(block.table_name),
                    })
                structured_data.append(
                    {
                        "type": "table",
                        "data": {
                            "headers": block.headers,
                            "rows": [row for row in block.rows],
                            "column_renderers": block.column_renderers,
                        },
                    }
                )
            elif isinstance(block, KubernetesDiffBlock):
                structured_data.append(
                    {
                        "type": "diff",
                        "data": {
                            "old": block.old,
                            "new": block.new,
                            "resource_name": block.resource_name,
                            "num_additions": block.num_additions,
                            "num_deletions": block.num_deletions,
                            "num_modifications": block.num_modifications,
                            "updated_paths": [d.formatted_path for d in block.diffs],
                        },
                    }
                )
            elif isinstance(block, CallbackBlock):
                callbacks = []
                for (text, callback) in block.choices.items():
                    callbacks.append(
                        {
                            "text": text,
                            "callback": ExternalActionRequestBuilder.create_for_func(
                                callback,
                                sink_name,
                                text,
                                account_id,
                                cluster_id,
                                signing_key,
                            ).json(),
                        }
                    )

                structured_data.append({"type": "callbacks", "data": callbacks})
            else:
                logging.error(
                    f"cannot convert block of type {type(block)} to robusta platform format block: {block}"
                )
                continue  # no reason to crash the entire report

        return {
            "issue_id": str(finding_id),
            "file_type": "structured_data",
            "data": json.dumps(structured_data),
            "account_id": account_id,
        }





from .base import *
from .blocks import *

__all__ = [
    "BaseBlock",
    "MarkdownBlock",
    "DividerBlock",
    "FileBlock",
    "HeaderBlock",
    "ListBlock",
    "TableBlock",
    "KubernetesFieldsBlock",
    "CallbackBlock",
]
